"""Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import enum
import nanosurf.lib.devices.i2c.bus_master as i2c 

class RampStatFlags(enum.IntEnum):
    Stop_Left             = 0x00000001
    Stop_Right            = 0x00000002
    Stop_Left_Latched     = 0x00000004
    Stop_Right_Latched    = 0x00000008
    Stop_Left_Event       = 0x00000010
    Stop_Right_Event      = 0x00000020
    Stop_StallGuard_Event = 0x00000040
    Pos_Reached_Event     = 0x00000080
    Velocity_Reached      = 0x00000100
    TargetPos_Reached     = 0x00000200
    ZeroVelocity_Reached  = 0x00000400
    T_Wait_Active         = 0x00000800
    Second_Move_Active    = 0x00001000
    Status_StallGuard     = 0x00002000

class SWModeFlags(enum.IntEnum):
    Switch_Left_Enabled     = 0x00000001
    Switch_Right_Enabled    = 0x00000002
    Switch_Left_HighActive  = 0x00000004
    Switch_Right_HighActive = 0x00000008
    Swap_Left_Right_Switch  = 0x00000010
    LatchPos_Left_Active    = 0x00000020
    LatchPos_Left_Inactive  = 0x00000040
    LatchPos_Right_Active   = 0x00000080
    LatchPos_Right_Inactive = 0x00000100
    Reserved                = 0x00000200
    StallGuard_Stop_Enabled = 0x00000400
    SoftStop_Enabled        = 0x00000800

class Reg_Channel(enum.IntEnum):
    # General Configuration
    GCONF = 0
    GSTAT = 1
    TESTSEL = 2
    INPUT = 3
    X_COMPARE= 4
    # RampGen Motion Control
    RAMPMODE =  5
    XACTUAL =  6
    VACTUAL =  7
    VSTART =  8
    A1 =  9
    V1 =  10
    AMAX =  11
    VMAX =  12
    DMAX =  13
    D1 =  14
    VSTOP =  15
    TZEROWAIT = 16
    XTARGET =  17
    # RampGen Driver Features Control
    IHOLD_IRUN =  18
    VCOOLTHRS =  19
    VHIGH =  20
    SW_MODE =  21
    RAMP_STAT =  22
    XLATCH =  23
    # Motor Drive Registers
    MSLUT_0 =  24
    MSLUT_1 =  25
    MSLUT_2 =  26
    MSLUT_3 =  27
    MSLUT_4 =  28
    MSLUT_5 =  29
    MSLUT_6 =  30
    MSLUT_7 =  31
    MSLUTSEL =  32
    MSLUTSTART =  33
    MSCNT =  34
    MSCURACT =  35
    CHOPCONF =  36
    COOLCONF =  37
    DRV_STATUS =  38

class Chip_TMC5031(i2c.I2CChip): 
    """ stepper motor controller chip
    connected to I2C by a I2C<->SPI bridge 
    """

    class SPI_Function_IDS(enum.IntEnum):
        slave_select_0  = 0x01
        slave_select_1  = 0x02
        slave_select_2  = 0x04
        slave_select_3  = 0x08
        slave_select_23 = 0x04 | 0x08
        config_spi      = 0xf0
        clear_int       = 0xf1
        idle_mode       = 0xf2
        gpio_write      = 0xf4
        gpio_read       = 0xf5
        gpio_enable     = 0xf6
        gpio_config     = 0xf7

    class Reg_TMC(enum.IntEnum):
        #// General Configuration
        GCONF = 0x00
        GSTAT = 0x01
        TESTSEL = 0x03
        INPUT = 0x04
        X_COMPARE = 0x05
        # RampGen Motion Control
        M1_RAMPMODE = 0x20
        M1_XACTUAL = 0x21
        M1_VACTUAL = 0x22
        M1_VSTART = 0x23
        M1_A1 = 0x24
        M1_V1 = 0x25
        M1_AMAX = 0x26
        M1_VMAX = 0x27
        M1_DMAX = 0x28
        M1_D1 = 0x2A
        M1_VSTOP = 0x2B
        M1_TZEROWAIT = 0x2C
        M1_XTARGET = 0x2D
        M2_RAMPMODE = 0x40
        M2_XACTUAL = 0x41
        M2_VACTUAL = 0x42
        M2_VSTART = 0x43
        M2_A1 = 0x44
        M2_V1 = 0x45
        M2_AMAX = 0x46
        M2_VMAX = 0x47
        M2_DMAX = 0x48
        M2_D1 = 0x4A
        M2_VSTOP = 0x4B
        M2_TZEROWAIT = 0x4C
        M2_XTARGET = 0x4D
        # RampGen Driver Features Control
        M1_IHOLD_IRUN = 0x30
        M1_VCOOLTHRS = 0x31
        M1_VHIGH = 0x32
        M1_SW_MODE = 0x34
        M1_RAMP_STAT = 0x35
        M1_XLATCH = 0x36
        M2_IHOLD_IRUN = 0x50
        M2_VCOOLTHRS = 0x51
        M2_VHIGH = 0x52
        M2_SW_MODE = 0x54
        M2_RAMP_STAT = 0x55
        M2_XLATCH = 0x56
        # Motor Drive Registers
        M1_MSLUT_0 = 0x60
        M1_MSLUT_1 = 0x61
        M1_MSLUT_2 = 0x62
        M1_MSLUT_3 = 0x63
        M1_MSLUT_4 = 0x64
        M1_MSLUT_5 = 0x65
        M1_MSLUT_6 = 0x66
        M1_MSLUT_7 = 0x67
        M1_MSLUTSEL = 0x68
        M1_MSLUTSTART = 0x69
        M1_MSCNT = 0x6A
        M1_MSCURACT = 0x6B
        M1_CHOPCONF = 0x6C
        M1_COOLCONF = 0x6D
        M1_DRV_STATUS = 0x6F
        M2_MSLUT_0 = 0x70
        M2_MSLUT_1 = 0x71
        M2_MSLUT_2 = 0x72
        M2_MSLUT_3 = 0x73
        M2_MSLUT_4 = 0x74
        M2_MSLUT_5 = 0x75
        M2_MSLUT_6 = 0x76
        M2_MSLUT_7 = 0x77
        M2_MSLUTSEL = 0x78
        M2_MSLUTSTART = 0x79
        M2_MSCNT = 0x7A
        M2_MSCURACT = 0x7B
        M2_CHOPCONF = 0x7C
        M2_COOLCONF = 0x7D
        M2_DRV_STATUS = 0x7F
    
    config_set_first:dict[Reg_Channel, Reg_TMC] = {
        Reg_Channel.GCONF: Reg_TMC.GCONF,
        Reg_Channel.GSTAT: Reg_TMC.GSTAT,
        Reg_Channel.TESTSEL: Reg_TMC.TESTSEL,
        Reg_Channel.INPUT: Reg_TMC.INPUT,
        Reg_Channel.X_COMPARE: Reg_TMC.X_COMPARE,
        Reg_Channel.RAMPMODE: Reg_TMC.M1_RAMPMODE,
        Reg_Channel.XACTUAL: Reg_TMC.M1_XACTUAL,
        Reg_Channel.VACTUAL: Reg_TMC.M1_VACTUAL,
        Reg_Channel.VSTART: Reg_TMC.M1_VSTART,
        Reg_Channel.A1: Reg_TMC.M1_A1,
        Reg_Channel.V1: Reg_TMC.M1_V1,
        Reg_Channel.AMAX: Reg_TMC.M1_AMAX,
        Reg_Channel.VMAX: Reg_TMC.M1_VMAX,
        Reg_Channel.DMAX: Reg_TMC.M1_DMAX,
        Reg_Channel.D1: Reg_TMC.M1_D1,
        Reg_Channel.VSTOP: Reg_TMC.M1_VSTOP,
        Reg_Channel.TZEROWAIT: Reg_TMC.M1_TZEROWAIT,
        Reg_Channel.XTARGET: Reg_TMC.M1_XTARGET,
        Reg_Channel.IHOLD_IRUN: Reg_TMC.M1_IHOLD_IRUN,
        Reg_Channel.VCOOLTHRS: Reg_TMC.M1_VCOOLTHRS,
        Reg_Channel.VHIGH: Reg_TMC.M1_VHIGH,
        Reg_Channel.SW_MODE: Reg_TMC.M1_SW_MODE,
        Reg_Channel.RAMP_STAT: Reg_TMC.M1_RAMP_STAT,
        Reg_Channel.XLATCH: Reg_TMC.M1_XLATCH,
        Reg_Channel.MSLUT_0: Reg_TMC.M1_MSLUT_0,
        Reg_Channel.MSLUT_1: Reg_TMC.M1_MSLUT_1,
        Reg_Channel.MSLUT_2: Reg_TMC.M1_MSLUT_2,
        Reg_Channel.MSLUT_3: Reg_TMC.M1_MSLUT_3,
        Reg_Channel.MSLUT_4: Reg_TMC.M1_MSLUT_4,
        Reg_Channel.MSLUT_5: Reg_TMC.M1_MSLUT_5,
        Reg_Channel.MSLUT_6: Reg_TMC.M1_MSLUT_6,
        Reg_Channel.MSLUT_7: Reg_TMC.M1_MSLUT_7,
        Reg_Channel.MSLUTSEL: Reg_TMC.M1_MSLUTSEL,
        Reg_Channel.MSLUTSTART: Reg_TMC.M1_MSLUTSTART,
        Reg_Channel.MSCNT: Reg_TMC.M1_MSCNT,
        Reg_Channel.MSCURACT: Reg_TMC.M1_MSCURACT,       
        Reg_Channel.CHOPCONF: Reg_TMC.M1_CHOPCONF,
        Reg_Channel.COOLCONF: Reg_TMC.M1_COOLCONF,
        Reg_Channel.DRV_STATUS: Reg_TMC.M1_DRV_STATUS
    }

    config_set_second:dict[Reg_Channel, Reg_TMC] = {
        Reg_Channel.GCONF: Reg_TMC.GCONF,
        Reg_Channel.GSTAT: Reg_TMC.GSTAT,
        Reg_Channel.TESTSEL: Reg_TMC.TESTSEL,
        Reg_Channel.INPUT: Reg_TMC.INPUT,
        Reg_Channel.X_COMPARE: Reg_TMC.X_COMPARE,
        Reg_Channel.RAMPMODE: Reg_TMC.M2_RAMPMODE,
        Reg_Channel.XACTUAL: Reg_TMC.M2_XACTUAL,
        Reg_Channel.VACTUAL: Reg_TMC.M2_VACTUAL,
        Reg_Channel.VSTART: Reg_TMC.M2_VSTART,
        Reg_Channel.A1: Reg_TMC.M2_A1,
        Reg_Channel.V1: Reg_TMC.M2_V1,
        Reg_Channel.AMAX: Reg_TMC.M2_AMAX,
        Reg_Channel.VMAX: Reg_TMC.M2_VMAX,
        Reg_Channel.DMAX: Reg_TMC.M2_DMAX,
        Reg_Channel.D1: Reg_TMC.M2_D1,
        Reg_Channel.VSTOP: Reg_TMC.M2_VSTOP,
        Reg_Channel.TZEROWAIT: Reg_TMC.M2_TZEROWAIT,
        Reg_Channel.XTARGET: Reg_TMC.M2_XTARGET,
        Reg_Channel.IHOLD_IRUN: Reg_TMC.M2_IHOLD_IRUN,
        Reg_Channel.VCOOLTHRS: Reg_TMC.M2_VCOOLTHRS,
        Reg_Channel.VHIGH: Reg_TMC.M2_VHIGH,
        Reg_Channel.SW_MODE: Reg_TMC.M2_SW_MODE,
        Reg_Channel.RAMP_STAT: Reg_TMC.M2_RAMP_STAT,
        Reg_Channel.XLATCH: Reg_TMC.M2_XLATCH,
        Reg_Channel.MSLUT_0: Reg_TMC.M2_MSLUT_0,
        Reg_Channel.MSLUT_1: Reg_TMC.M2_MSLUT_1,
        Reg_Channel.MSLUT_2: Reg_TMC.M2_MSLUT_2,
        Reg_Channel.MSLUT_3: Reg_TMC.M2_MSLUT_3,
        Reg_Channel.MSLUT_4: Reg_TMC.M2_MSLUT_4,
        Reg_Channel.MSLUT_5: Reg_TMC.M2_MSLUT_5,
        Reg_Channel.MSLUT_6: Reg_TMC.M2_MSLUT_6,
        Reg_Channel.MSLUT_7: Reg_TMC.M2_MSLUT_7,
        Reg_Channel.MSLUTSEL: Reg_TMC.M2_MSLUTSEL,
        Reg_Channel.MSLUTSTART: Reg_TMC.M2_MSLUTSTART,
        Reg_Channel.MSCNT: Reg_TMC.M2_MSCNT, 
        Reg_Channel.MSCURACT: Reg_TMC.M2_MSCURACT,       
        Reg_Channel.CHOPCONF: Reg_TMC.M2_CHOPCONF,
        Reg_Channel.COOLCONF: Reg_TMC.M2_COOLCONF,
        Reg_Channel.DRV_STATUS: Reg_TMC.M2_DRV_STATUS
    }

    class Channels(enum.IntEnum):
        first = 0
        second = 1

    class TMC_Chip(enum.IntEnum):
        first = 0
        second = 1
    
    def __init__(self, bus_addr: int, **kwargs):
        super().__init__(bus_addr, offset_mode=i2c.I2COffsetMode.NoOffset, auto_lock=True, **kwargs)
        self._spi_gpio_cache = 0
        self._last_spi_status = 0

    def init_spi_bridge(self):    
        self.spi_bridge_config_spi(0x0c)  # 0x0c = ORDER=MSB first,  MODE0/1=CPHA=1, CPOL=1, F0/1(SPI_Clk)=1.8MHz
        self.spi_bridge_gpio_config(0x50) #  eDrvEnSlave(SlaveSelect2 + SlaveSelect3) = Push-Pull output
        self.spi_bridge_gpio_enable(Chip_TMC5031.SPI_Function_IDS.config_spi.slave_select_23) # activate DrvEn-Signal as GPIO 
        self.spi_bridge_gpio_write(0xff)  # Disable all (1 = off, 0 = on)

    def spi_bridge_write_byte(self, id: SPI_Function_IDS, val: int):
        self.write_bytes([id, val])

    def spi_bridge_read_byte(self, id: SPI_Function_IDS) -> int:
        self.write_byte(id)
        return self.read_byte()

    def spi_bridge_config_spi(self, val: int):
        self.spi_bridge_write_byte(Chip_TMC5031.SPI_Function_IDS.config_spi, val)

    def spi_bridge_clear_int(self):
        self.spi_bridge_write_byte(Chip_TMC5031.SPI_Function_IDS.clear_int, 0)

    def spi_bridge_set_idle_mode(self):
        self.spi_bridge_write_byte(Chip_TMC5031.SPI_Function_IDS.idle_mode, 0)

    def spi_bridge_gpio_write(self, val: int):
        self._spi_gpio_cache = val
        self.spi_bridge_write_byte(Chip_TMC5031.SPI_Function_IDS.gpio_write, self._spi_gpio_cache)

    def spi_bridge_gpio_read(self) -> int:
        return self.spi_bridge_read_byte(Chip_TMC5031.SPI_Function_IDS.gpio_read)

    def spi_bridge_gpio_enable(self, val: int):
        self.spi_bridge_write_byte(Chip_TMC5031.SPI_Function_IDS.gpio_enable, val)

    def spi_bridge_gpio_config(self, val: int):
        self.spi_bridge_write_byte(Chip_TMC5031.SPI_Function_IDS.gpio_config, val)

    def spi_bridge_write_buffer(self, id: SPI_Function_IDS, data: list[int]):
        buffer = [int(id)]
        buffer.extend(data)
        self.write_bytes(buffer)

    def spi_bridge_read_buffer(self, size: int) -> list[int]:
        return self.read_bytes(size)

    def motor_to_chip_ch(self, motor: int) -> tuple:
        chip = int(motor / 2)
        ch   = int(motor % 2)
        return (chip, ch)        

    def channel_to_tmc_reg(self, ch: Channels, reg: Reg_Channel) -> Reg_TMC:
        if ch == Chip_TMC5031.Channels.first:
            return Chip_TMC5031.config_set_first[reg]
        elif ch == Chip_TMC5031.Channels.second:
            return Chip_TMC5031.config_set_second[reg]
        assert False, f"Illegal channel number {ch}"

    def chip_to_spi_function_register(self, chip: TMC_Chip) -> SPI_Function_IDS:
        if chip == Chip_TMC5031.TMC_Chip.first:
            return Chip_TMC5031.SPI_Function_IDS.slave_select_0
        elif chip == Chip_TMC5031.TMC_Chip.second:
            return Chip_TMC5031.SPI_Function_IDS.slave_select_1
        assert False, f"Illegal chip number {chip}"

    def motor_driver_enable(self, motor: int, enable: bool):
        chip, ch = self.motor_to_chip_ch(motor)
        if  chip == Chip_TMC5031.TMC_Chip.first:
            id = Chip_TMC5031.SPI_Function_IDS.slave_select_2
        elif chip == Chip_TMC5031.TMC_Chip.second:
            id = Chip_TMC5031.SPI_Function_IDS.slave_select_3
        else:
            assert False, f"Illegal chip number {chip}"

        if enable:
            self.spi_bridge_gpio_write(self._spi_gpio_cache & (~id))
        else:
            self.spi_bridge_gpio_write(self._spi_gpio_cache | id)

    def is_motor_driver_enabled(self, motor: int) -> bool:
        chip, ch = self.motor_to_chip_ch(motor)
        if chip == Chip_TMC5031.TMC_Chip.first:
            id = Chip_TMC5031.SPI_Function_IDS.slave_select_2
        elif chip == Chip_TMC5031.TMC_Chip.second:
            id = Chip_TMC5031.SPI_Function_IDS.slave_select_3
        else:
            assert False, f"Illegal chip number {chip}"
        
        val = self.spi_bridge_gpio_read()
        return (val & (~id)) > 0

    def write_register(self, motor: int, reg: Reg_Channel, val: int, signed: bool = False): 
        chip, ch = self.motor_to_chip_ch(motor)
        buffer:list[int] = [self.channel_to_tmc_reg(ch, reg) | 0x80] # write access
        try:
            buffer.extend(val.to_bytes(4, byteorder='big', signed=signed))
            self.spi_bridge_write_buffer(self.chip_to_spi_function_register(chip), buffer)
        except OverflowError:
            print(f"value out of range error: '{val}'")

    def read_register(self, motor: int, reg: Reg_Channel, signed: bool = False) -> int:
        chip, ch = self.motor_to_chip_ch(motor)
        buffer:list[int] = [self.channel_to_tmc_reg(ch, reg)] # read access
        buffer.extend([0,0,0,0])

        # have to send it twice in order to get data from selected register, see TMC5031 data sheet for read access
        self.spi_bridge_write_buffer(self.chip_to_spi_function_register(chip), buffer)
        self.spi_bridge_write_buffer(self.chip_to_spi_function_register(chip), buffer)

        
        read_bytes = 5
        read_buffer = self.spi_bridge_read_buffer(read_bytes)
        assert len(read_buffer) == 5, f"Register read error. Assumed {read_bytes} bytes, but got {len(read_buffer)} bytes."
 
        self._last_spi_status = read_buffer[0]
        return int.from_bytes(read_buffer[1:], byteorder='big', signed=signed)

    def get_last_status(self) -> int:
        return self._last_spi_status




""" Table Widget enhancement to simplify "key=value" style tables
Copyright Nanosurf AG 2021
License - MIT
"""

import enum
import typing
from PySide2 import QtWidgets, QtCore
import nanosurf.lib.datatypes.sci_val as sci_val

class TableEntryIDs(enum.IntEnum):
   """
   Subclass this enum class and define a identifier for each table entry
   First entry must have index 0. eg MyFirstVal = 0
   Then define each TableEntryID by NSFNameValueTable.define_entry() during gui preparation
   """
   pass

class NSFNameValueTable(QtWidgets.QTableWidget):
   def __init__(self, table_entry_ids: TableEntryIDs, parent: QtWidgets.QWidget = None) -> None:
      super().__init__()
      self.setEnabled(False)
      self.verticalHeader().hide()
      self.horizontalHeader().hide()
      self.horizontalHeader().setDefaultSectionSize(100)
      self.horizontalHeader().setStretchLastSection(True)
      self.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
      self.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
      self.setColumnCount(2)
      self.setHorizontalHeaderLabels(["Name","Value"])
      self.define_table(table_entry_ids)

   def define_table(self, table_entry_ids:TableEntryIDs):
      self._table_ids = table_entry_ids
      self.setRowCount(len(list(self._table_ids)))
      self.setMinimumSize(0,(self.rowCount()+1)*30)  

   def define_entry(self, entry_id:TableEntryIDs, name:str, value:str="-") -> None:
      self.setItem(entry_id,0,QtWidgets.QTableWidgetItem(name))
      self.setItem(entry_id,1,QtWidgets.QTableWidgetItem(value))

   def clear_values(self, value_default:str = "-") -> None:
      for id in self._table_ids:
         self.item(id, 1).setText(value_default)

   def clear_value(self, entry_id:TableEntryIDs, value_default:str = "-") -> None:
      self.item(entry_id,1).setText(value_default)

   def get_entry_count(self):
      return len(list(self._table_ids))

   def set_value(self, entry_id:TableEntryIDs, value, unit="", precision=3):
      str_val = ""
      if isinstance(value, sci_val.SciVal):
         str_val = value.to_string_formatted(sci_val.up.Prefix.auto_ ,precision=precision)
      elif isinstance(value, str):
         str_val = value
      else:
         val = sci_val.SciVal(value, unit_str=unit)
         str_val = val.to_string_formatted(sci_val.up.Prefix.auto_, precision=precision)
      self.item(entry_id,1).setText(str_val)

   def get_value(self, entry_id:TableEntryIDs) ->sci_val.SciVal:
        val_str = self.item(entry_id,1).text()
        res = sci_val.convert.to_value(val_str)
        if res.success:
            return sci_val.SciVal(res.value, res.unit)
        return sci_val.SciVal(0.0,"")

class NSFTable(QtWidgets.QTableWidget):
   def __init__(self, header: list[str], initial_entries=1, hide_header: bool = False, parent: QtWidgets.QWidget = None) -> None:
      super().__init__()
      self.setEnabled(False)
      self.verticalHeader().hide()
      if hide_header: self.horizontalHeader().hide()
      self.horizontalHeader().setDefaultSectionSize(100)
      self.horizontalHeader().setStretchLastSection(True)
      self.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
      self.horizontalHeader().setDefaultAlignment(QtCore.Qt.AlignLeft)
      self.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
      self.setColumnCount(len(header))
      self.setHorizontalHeaderLabels(header)
      self.define_table(initial_entries)
      self.setMinimumSize(0,(self.rowCount()+1)*30)  

   def define_table(self, entries: int):
      current_len = self.rowCount()
      items = ["" for col in range(self.columnCount())]
      self.setRowCount(entries)
      for new_entry in range(current_len, entries):
         self.define_entry(new_entry, items)

   def define_entry(self, entry_id: int, items: list[str], flags: int = QtCore.Qt.ItemIsSelectable) -> None:
      for col_index, value in enumerate(items):
         self.setItem(entry_id, col_index, QtWidgets.QTableWidgetItem(value))
         self.item(entry_id, col_index).setFlags(flags)

   def clear_values(self, value_default:str = "") -> None:
      for index in range(self.rowCount()):
         self.clear_value(index, value_default)

   def clear_value(self, entry_id: int, value_default:str = "") -> None:
      for col_index in range(self.columnCount()):
         self.item(entry_id, col_index).setText(value_default)

   def set_entry_flags(self, entry_id: int, col_index: int, flags: int) -> None:
      self.item(entry_id, col_index).setFlags(flags)

   def set_column_flags(self, col_index: int, flags: int) -> None:
      for row_index in range(self.rowCount()):
         self.item(row_index, col_index).setFlags(flags)
         if (flags &  QtCore.Qt.ItemIsUserCheckable) > 0:
            self.item(row_index, col_index).setCheckState(QtCore.Qt.CheckState.Unchecked)

   def get_entry_count(self):
      return self.rowCount()

   def set_value(self, entry_id:TableEntryIDs, col_pos: int, value: typing.Union[str, sci_val.SciVal, int, float], unit: str = "", precision: int = 3):
      str_val = ""
      if isinstance(value, sci_val.SciVal):
         str_val = value.to_string_formatted(sci_val.up.Prefix.auto_ ,precision=precision)
      elif isinstance(value, str):
         str_val = value
      else:
         val = sci_val.SciVal(value, unit_str=unit)
         str_val = val.to_string_formatted(sci_val.up.Prefix.auto_, precision=precision)
      self.item(entry_id,col_pos).setText(str_val)

   def get_value(self, entry_id: int, col_pos: int) -> str:
        val_str = self.item(entry_id,col_pos).text()
        return val_str



"""Package for scripting the Nansurf control software.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

from enum import IntEnum

class DAC(IntEnum):
    HiResOut_POSITIONX = 1
    HiResOut_POSITIONY = 2
    HiResOut_POSITIONZ = 3
    HiResOut_POSITIONW = 4
    HiResOut_TIPVOLTAGE = 5
    HiResOut_OUT6 = 6
    HiResOut_OUT7 = 7
    HiResOut_OUT8 = 8
    HiResOut_USER1= 9
    HiResOut_USER2 = 10
    HiResOut_USER3 = 11
    HiResOut_USER4 = 12
    FastOut_EXCITATION = 13
    FastOut_USER = 14
    FastOut_FAST2 = 15
    FastOut_FAST3 = 16

class ADC(IntEnum):
    HiResIn_DEFLECTION = 1
    HiResIn_LATERAL = 2
    HiResIn_POSITIONX = 3
    HiResIn_POSITIONY = 4
    HiResIn_POSITIONZ = 5
    HiResIn_DETECTORSUM = 6
    HiResIn_TIPCURRENT = 7
    HiResIn_IN6 = 8
    HiResIn_USER1 = 9
    HiResIn_USER2 = 10
    HiResIn_USER3 = 11
    HiResIn_USER4 = 12
    FastIn_DEFLECTION = 13
    FastIn_CH2 = 14
    FastIn_USER = 15

class Analyzer(IntEnum):
    SigAnalyzer_1 = 1
    SigAnalyzer_2 = 2

class ChannelMux(IntEnum):
    AnalogHiResOut = 1
    AnalogFastOut = 2
    SigAnalyzer_1 = 3
    SigAnalyzer_2 = 4

class DataCapture(IntEnum):
    CaptureHiRes = 5
    CaptureFast = 6

class DataSampler(IntEnum):
    SamplerHiRes = 7


class _CtrlUnit:
    def __init__(self, spm):
        self._spm = spm
        self._lu = None
        self.__source = None
        self.__target = None

    @property
    def source(self):
        return self.__source

    @source.setter
    def source(self, val):
        self.__source = val

    @property
    def target(self):
        return self.__target

    @target.setter
    def target(self, val):
        self.__target = val




""" This is a file with auxiliary functions for GUI template using Nanosurf Studio Style
Copyright Nanosurf AG 2021
License - MIT
"""
import pathlib
import pyqtgraph.exporters

# Save data in multiple columns: data[channel][data] fist dimension = channels, second dimension = data stream
def savedata_txt(file_name: pathlib.Path, data, header:str="", separator="\t"):
    with open(file_name, 'w+') as f:
        if header != "":
            f.write(header+"\n")
        for i in range(len(data[0])):
            write_string = ""
            for j in range(len(data)):
                if j == len(data)-1:
                    write_string =  write_string + str(data[j][i]) + "\n"
                else:
                    write_string =  write_string + str(data[j][i]) + separator
            f.write(write_string)
        f.close()

# Load data
def loaddata_txt(file_name: pathlib.Path, vertical = True, skip_header: int = 0, separator="\t"):
    result = []
    with open(file_name, 'r') as f:
        lines = f.readlines()
        for line in lines:
            if skip_header <= 0:
                sline = line.rstrip()
                splitline = sline.split(separator)
                floatline = [float(i) for i in splitline]
                result.append(floatline)
            else:
                skip_header -= 1
    if vertical:
        result = list(map(list, zip(*result)))
    return result


def saveplot_png(file_name: pathlib.Path, plotitem):
    exporter = pyqtgraph.exporters.ImageExporter(plotitem)
    #exporter.parameters()['width'] = 100   # (note this also affects height parameter)
    file_name.with_suffix('.png')
    print(file_name)
    exporter.export(str(file_name))
    return file_name

def save_results(file_name: pathlib.Path, resulttable):    
    with open(file_name, 'w+') as f:
        for id in range(resulttable.rowCount()):
            name_str = resulttable.item(id,0).text()
            val_str = resulttable.item(id,1).text()
            result_string = name_str + "; "+ val_str +("\n")
            f.write(result_string)
    f.close()



""" The long lasting worker thread as demo - just wait some time and create data periodically
Copyright Nanosurf AG 2021
License - MIT
"""
import time
import enum
import pythoncom # to connect to SPM Controller within thread
from PySide2.QtCore import Signal
import nanosurf
from nanosurf.lib.datatypes import nsf_thread, sci_channel, sci_stream
import nanosurf.lib.spm.lowlevel as ll
import nanosurf.lib.spm.lowlevel.ctrlunits as cu
from app import app_common, module_base
from modules.switching_spec_module import settings

class SpecWorkerData():
    def __init__(self):
        self.output : float = 0.0
        self.output_unit = str = ""
        self.amplitude_on : float = 0.0
        self.amplitude_off : float = 0.0
        self.amplitude_unit = str = ""
        self.phase_on : float = 0.0
        self.phase_off : float = 0.0

OutputChannel_ID_to_LUInst_Map = {
    settings.OutputChannelID.User1.value : cu.DAC.HiResOut_USER1,
    settings.OutputChannelID.User2.value : cu.DAC.HiResOut_USER2,
    settings.OutputChannelID.TipVoltage.value : cu.DAC.HiResOut_TIPVOLTAGE,
}

class SpecDir(enum.Enum):
    SpecFwd = 0
    SpecBwd = 1

class SpecWorker(nsf_thread.NSFBackgroundWorker):
    """ This class implements the long lasting activity in the background to not freeze the gui """

    sig_tick = Signal() # is emitted if par_send_tick is True
    sig_new_data = Signal() # is emitted each time new data are available . Result is read by self.get_result()

    """ parameter for the background work"""
    par_output_id = settings.OutputChannelID.User1
    par_output_span = 10.0   # V: spec range, using the scaling and unit as defined in the software
    par_output_center = 0.0
    par_time_per_step = 0.1    # equilibration time for voltage and lock-in to settle
    par_steps = 10             # number of steps per direction

    _sig_message = Signal(str, int)

    def __init__(self, my_module: module_base.ModuleBase):
        self.module = my_module
        self.resulting_data = SpecWorkerData()
        self.spm:nanosurf.Spm = None
        super().__init__()
        self._sig_message.connect(self.module.app.show_message)

    def do_work(self):
        """ This is the working function for the long task"""
        # clear data
        self.resulting_data = SpecWorkerData()

        if self._connect_to_controller():
                           
            lockin = self.spm.lowlevel.ctrlunits.get_lock_in()
            dacout = self.spm.lowlevel.ctrlunits.get_dac(OutputChannel_ID_to_LUInst_Map[self.par_output_id])
            original_out = dacout.dc

            self.cur_voltage_range = self.par_output_span
            self.cur_steps = self.par_steps 
            self.cur_time_per_step = self.par_time_per_step 
            self.cur_offset_voltage = self.par_output_center

            self.resulting_data.amplitude_unit = lockin.input_amp_unit 
            self.resulting_data.output_unit = dacout.unit
            
            #move in both spectroscopy directions
            for spec_dir in SpecDir:
                #voltage step size with opposite sign for fwd and bwd
                cur_offset = self.cur_offset_voltage  + self.cur_voltage_range/2*(2*(spec_dir.value%2)-1)
                cur_voltage_step = self.cur_voltage_range/(self.cur_steps-1)*(1-2*(spec_dir.value%2)) 
                
                # This loop creates the shape of switching spectroscopy in time
                for cur_step in range(self.cur_steps): 
                    cur_output = cur_offset + cur_step*cur_voltage_step
                    self.resulting_data.output = cur_output

                    # write a dc level to an output and measure response during voltage application
                    dacout.dc = cur_output
                    time.sleep(self.cur_time_per_step)
                    self.resulting_data.amplitude_on = lockin.input_amp 
                    self.resulting_data.phase_on = lockin.input_phase

                    # switch potential back to offset voltage and measure response again
                    dacout.dc = self.cur_offset_voltage 
                    time.sleep(self.cur_time_per_step)
                    self.resulting_data.amplitude_off = lockin.input_amp 
                    self.resulting_data.phase_off = lockin.input_phase
            
                    if self.is_stop_request_pending():
                        break
                    self.sig_new_data.emit()

            # restore original setting of output        
            dacout.dc = original_out
        self._disconnect_from_controller()

    def get_result(self) -> SpecWorkerData:
        return self.resulting_data

    def send_message(self, msg:str, msg_type: app_common.MsgType = app_common.MsgType.Info):
        self._sig_message.emit(msg, msg_type)
        self.logger.info(msg)   

    def _connect_to_controller(self) -> bool:
        self.send_message("Connecting to Nanosurf controller")
        pythoncom.CoInitialize()
        ok = False
        if self.spm is None:
            self.spm = nanosurf.SPM()
            if self.spm.is_connected():
                if self.spm.is_scripting_enabled():
                    ok = True
                else:
                    self.send_message("Error: Scripting interface is not enabled", app_common.MsgType.Error)
            else:
                self.send_message("Error: Could not connect to controller. Check if software is started", app_common.MsgType.Error)
        else:
            ok = True
        return ok

    def _disconnect_from_controller(self):
        if self.spm is not None:
            if self.spm.application is not None:
                del self.spm
            self.spm = None
        


""" This is the screen of the module
Copyright Nanosurf AG 2021
License - MIT
"""

import pathlib
from PySide2 import QtWidgets
from PySide2.QtCore import Qt
import nanosurf as nsf
from app import module_base
from modules.switching_spec_module import module, settings

class ResultTableID(nsf.gui.nsf_tables.TableEntryIDs):
    """ identifier id are used in a nsf_table widget"""
    Items = 0
    last_data = 1
    mean_value = 2

ComboOutputChannels = [
    nsf.gui.NSFComboEntry(settings.OutputChannelID.User1,'User 1'),
    nsf.gui.NSFComboEntry(settings.OutputChannelID.User2,'User 2'),
    nsf.gui.NSFComboEntry(settings.OutputChannelID.TipVoltage,'Tip Voltage'),
]

""" some useful lists of allowed prefixes used by nsf_sci_edit widgets"""
allowed_count_units = [nsf.sci_val.up.Prefix.base]
allowed_time_units = [nsf.sci_val.up.Prefix.base, nsf.sci_val.up.Prefix.milli]
allowed_meter_units = [nsf.sci_val.up.Prefix.milli, nsf.sci_val.up.Prefix.micro, nsf.sci_val.up.Prefix.nano]
allowed_volt_units = [nsf.sci_val.up.Prefix.base, nsf.sci_val.up.Prefix.milli, nsf.sci_val.up.Prefix.micro]

class SwitchingSpecScreen(module_base.ModuleScreen):
    def __init__(self):
        super().__init__()

    def do_setup_screen(self, module: module.SwitchingSpecModule):
        """ create here your gui with all controls and their layout"""
        self.module = module

        # left layout - main controls ------------------------------------------------------------

        self.combo_output_id = nsf.gui.NSFComboBox(ComboOutputChannels,"Output")

        self.scival_output_span = nsf.gui.NSFSciEdit("Voltage range")
        self.scival_output_span.set_allowed_prefix_ids(allowed_volt_units)
        self.scival_output_span.set_prefix_id(nsf.sci_val.up.Prefix.base)
        self.scival_output_span.set_precision(2)
        self.scival_output_span.set_value_min_max(0.0, 2000.0)

        self.scival_output_center = nsf.gui.NSFSciEdit("Center offset")
        self.scival_output_center.set_allowed_prefix_ids(allowed_volt_units)
        self.scival_output_center.set_prefix_id(nsf.sci_val.up.Prefix.base)
        self.scival_output_center.set_precision(2)
        self.scival_output_center.set_value_min_max(-1000.0, +1000.0)
        
        self.scival_num_steps = nsf.gui.NSFSciEdit("Number of steps")
        self.scival_num_steps.set_allowed_prefix_ids(allowed_meter_units)
        self.scival_num_steps.set_prefix_id(nsf.sci_val.up.Prefix.base)
        self.scival_num_steps.set_precision(0)
        self.scival_num_steps.set_value_min_max(2, 1000)

        self.scival_time_delay = nsf.gui.NSFSciEdit("Time delay after step")
        self.scival_time_delay.set_allowed_prefix_ids(allowed_time_units)
        self.scival_time_delay.set_prefix_id(nsf.sci_val.up.Prefix.base)
        self.scival_time_delay.set_precision(2)
        self.scival_time_delay.set_value_min_max(0.0, +1.0)
        
        self.scival_file_index = nsf.gui.NSFSciEdit("File index")
        self.scival_file_index.set_allowed_prefix_ids(allowed_count_units)
        self.scival_file_index.set_prefix_id(nsf.sci_val.up.Prefix.base)
        self.scival_file_index.set_precision(0)
        self.scival_file_index.set_value_min_max(1, 10000)

        self.check_show_on_data = QtWidgets.QCheckBox("Show 'on' data")

        self.edit_file_mask = nsf.gui.NSFEdit("File name mask")
        self.edit_file_path = nsf.gui.NSFEdit("Destination path")

        self.check_auto_save = QtWidgets.QCheckBox("Auto save data")
        self.button_start_stop = QtWidgets.QPushButton("Start")        
        self.button_browse = QtWidgets.QPushButton("Browse")
        self.button_export = QtWidgets.QPushButton("Save data")

        self.layout_left = QtWidgets.QVBoxLayout()
        self.layout_left.addWidget(self.combo_output_id)
        self.layout_left.addWidget(self.scival_output_span)
        self.layout_left.addWidget(self.scival_output_center)
        self.layout_left.addWidget(self.scival_num_steps)
        self.layout_left.addWidget(self.scival_time_delay)
        self.layout_left.addSpacing(10)
        self.layout_left.addWidget(self.check_show_on_data)
        self.layout_left.addStretch()
        self.layout_left.addWidget(self.button_start_stop)
        self.layout_left.addStretch()

        # mid layout - plots and result ---------------------------------------------------------
        self.chart_amp = nsf.gui.NSFChart(logmodex=False)
        self.chart_phase = nsf.gui.NSFChart(logmodex=False)

        self.layout_mid = QtWidgets.QVBoxLayout()
        self.layout_mid.addWidget(self.chart_amp)
        self.layout_mid.addWidget(self.chart_phase)
        self.layout_path = QtWidgets.QHBoxLayout()
        self.layout_path.addWidget(self.edit_file_path)
        self.layout_path.addWidget(self.button_browse,alignment=Qt.AlignBaseline)        
        self.layout_mid.addLayout(self.layout_path)
        self.layout_file = QtWidgets.QHBoxLayout()
        self.layout_file_mask = QtWidgets.QVBoxLayout()
        self.layout_file_mask.addWidget(self.edit_file_mask)  
        self.layout_file.addLayout(self.layout_file_mask)      
        self.layout_file.addWidget(self.scival_file_index,alignment=Qt.AlignVCenter)  
        self.layout_file.addWidget(self.check_auto_save,alignment=Qt.AlignVCenter)
        self.layout_file.addStretch()      
        self.layout_file.addWidget(self.button_export,alignment=Qt.AlignVCenter)
        self.layout_mid.addLayout(self.layout_file)


        # right layout - additional user inputs
        self.layout_right= QtWidgets.QVBoxLayout()

        # set GUI controls
        self.screen_layout = QtWidgets.QHBoxLayout()
        # stretch only plot area and keep controls fix in size
        self.screen_layout.addLayout(self.layout_left, 0)
        self.screen_layout.addLayout(self.layout_mid,  1) 
        self.screen_layout.addLayout(self.layout_right,0)
        self.setLayout(self.screen_layout)

        self.bind_gui_elements()
        self.init_plot()
        self.enter_gui_state_idle()

    def bind_gui_elements(self):
        """ connect all gui widgets to module settings or any other source """
        
        # binding ProVal to widgets ensure that they are alway in sync 
        nsf.gui.connect_to_property(self.combo_output_id, self.module.settings.output_id)
        nsf.gui.connect_to_property(self.scival_output_span, self.module.settings.output_span)
        nsf.gui.connect_to_property(self.scival_output_center, self.module.settings.output_center)
        nsf.gui.connect_to_property(self.scival_num_steps, self.module.settings.number_of_steps)
        nsf.gui.connect_to_property(self.scival_time_delay, self.module.settings.time_delay_after_step)
        nsf.gui.connect_to_property(self.edit_file_path, self.module.settings.folder_name)
        nsf.gui.connect_to_property(self.edit_file_mask, self.module.settings.file_name_mask)
        nsf.gui.connect_to_property(self.scival_file_index, self.module.settings.file_index)
        nsf.gui.connect_to_property(self.check_auto_save, self.module.settings.auto_save_data)
        nsf.gui.connect_to_property(self.check_show_on_data, self.module.settings.show_on_data_values)

        # buttons have to be connected separately
        self.button_start_stop.clicked.connect(self.on_button_start_stop_clicked)    
        self.button_export.clicked.connect(self.on_button_export_clicked)    
        self.button_browse.clicked.connect(self.on_button_browse_clicked)    
        
        # listen to signals from the core module to react and update the gui
        self.module.sig_work_start_requested.connect(self.enter_gui_state_wait)
        self.module.sig_work_stop_requested.connect(self.enter_gui_state_wait)
        self.module.sig_work_active.connect(self.enter_gui_state_active)
        self.module.sig_work_done.connect(self.enter_gui_state_idle)
        self.module.sig_new_data_available.connect(self.show_new_data)
        self.module.sig_data_invalid.connect(self.set_data_invalid)

        self.module.settings.show_on_data_values.sig_value_changed.connect(self.change_plot_view)
        self.module.settings.output_id.sig_value_changed.connect(self.init_plot)

    def init_plot(self):
        self.chart_amp.set_title(f"Amplitude")
        self.chart_amp.set_label(nsf.gui.NSFChart.Axis.bottom, f"Output '{self.combo_output_id.current_entry_name()}'")
        self.chart_amp.set_unit(nsf.gui.NSFChart.Axis.bottom, "")
        self.chart_amp.set_unit(nsf.gui.NSFChart.Axis.left, "")  
        self.chart_amp.set_label(nsf.gui.NSFChart.Axis.left, "Amplitude")
        self.chart_amp.set_unit(nsf.gui.NSFChart.Axis.left, "V")  
        self.chart_amp.clear_plots()
        self.chart_phase.set_title(f"Phase")
        self.chart_phase.set_label(nsf.gui.NSFChart.Axis.bottom, f"Output '{self.combo_output_id.current_entry_name()}'")
        self.chart_phase.set_unit(nsf.gui.NSFChart.Axis.bottom, "")
        self.chart_phase.set_label(nsf.gui.NSFChart.Axis.left, "Phase")
        self.chart_phase.set_unit(nsf.gui.NSFChart.Axis.left, "°")  
        self.chart_phase.clear_plots()        

    def on_button_start_stop_clicked(self):
        if self.module.is_worker_busy():
            self.module.stop_worker()
        else:
            self.module.start_worker()

    def on_button_export_clicked(self):
        self.module.save_data()            

    def on_button_browse_clicked(self):
        dlg = QtWidgets.QFileDialog()
        dlg.setFileMode(QtWidgets.QFileDialog.DirectoryOnly)
        dlg.setDirectory(str(self.module.settings.folder_name.value))
        if dlg.exec_():
            fileNames = dlg.selectedFiles()
            self.module.settings.folder_name.value = pathlib.Path(fileNames[0])

    def enter_gui_state_wait(self):
        self.set_parameter_widget_enable_state(enabled=False)
        self.start_stop_button_state(wait=True)

    def enter_gui_state_active(self):
        self.set_parameter_widget_enable_state(enabled=False)
        self.start_stop_button_state(wait=False, stop_state=self.module.is_worker_busy())
        self.init_plot()

    def enter_gui_state_idle(self):
        self.set_parameter_widget_enable_state(enabled=True)
        self.start_stop_button_state(wait=False, stop_state=self.module.is_worker_busy())

    def set_parameter_widget_enable_state(self, enabled: bool = True):
        self.scival_output_span.setEnabled(enabled)
        self.scival_output_center.setEnabled(enabled)
        self.scival_num_steps.setEnabled(enabled)
        self.scival_time_delay.setEnabled(enabled)

    def start_stop_button_state(self, wait: bool = False, stop_state: bool = False):
        if wait:
            self.button_start_stop.setEnabled(False)
            self.button_start_stop.setText("Wait...")
        else:
            self.button_start_stop.setEnabled(True)
            self.button_start_stop.setText("Stop" if stop_state else "Start")

    def show_new_data(self):
        self.update_plot()
        self.update_result()

    def set_data_invalid(self):
        half_range = self.module.settings.output_span.value/2.0
        min_v = self.module.settings.output_center.value - half_range
        max_v = self.module.settings.output_center.value + half_range
        self.chart_amp.clear_plots()
        self.chart_amp.plot.setXRange(min_v, max_v)
        self.chart_phase.clear_plots()
        self.chart_phase.plot.setXRange(min_v, max_v)

    def update_result(self):
        res = self.module.get_result()
        # self.tableResults.set_value(ResultTableID.Items, res.number_of_data_points, "", precision=0)
        # self.tableResults.set_value(ResultTableID.last_data, res.last_data, "")
        # self.tableResults.set_value(ResultTableID.mean_value, res.mean_value, "")

    def update_plot(self): 
        current_data = self.module.get_result()
        self.chart_amp.set_unit(nsf.gui.NSFChart.Axis.bottom, current_data.outputs_unit)
        self.chart_phase.set_unit(nsf.gui.NSFChart.Axis.bottom, current_data.outputs_unit)  
        self.chart_amp.set_unit(nsf.gui.NSFChart.Axis.left, current_data.amplitudes_unit)  
        self.chart_amp.plot_data(x=current_data.outputs, y=current_data.amplitudes_off, layer_index=0)
        self.chart_phase.plot_data(x=current_data.outputs, y=current_data.phases_off, layer_index=0)
        if self.module.settings.show_on_data_values.value:
            self.chart_amp.plot_data(x=current_data.outputs, y=current_data.amplitudes_on, layer_index=1)
            self.chart_phase.plot_data(x=current_data.outputs, y=current_data.phases_on, layer_index=1)

    def change_plot_view(self):
        self.init_plot()
        self.update_plot()


""" The application base class of the application framework
Copyright Nanosurf AG 2021
License - MIT
"""

import sys
import os
import ctypes
import logging
import pathlib
from PySide2.QtWidgets import QApplication
from PySide2.QtCore import QSettings

from app import module_base
from app import app_common
from app import app_gui

import nanosurf.lib.util.fileutil as fileutil
import nanosurf.lib.datatypes.prop_val as prop_val

dir_name_log = "log"
dir_name_config = "config"
dir_name_app = "app"
dir_name_resources = "app"

class AppSettings(prop_val.PropStore):
    Logging  = prop_val.PropVal(True)
    LoadLastSettings = prop_val.PropVal(True)
    ActiveModuleIndex = prop_val.PropVal(int(0))

class ApplicationBase(QApplication):
    def __init__(self, company: str, app_name_short: str, app_name_long: str,  *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.company = company
        self.app_name = app_name_short
        self.app_name_long = app_name_long 
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(app_name_short)
        self.is_in_debugging_mode = (getattr(sys, 'gettrace', lambda : None)() is not None)

        """ Standard registry connection is defined here for further usage also by modules"""
        self.registry = QSettings(self.company, self.app_name)

        """ Standard files and folders are defined here for further usage also by modules"""
        self.main_path = pathlib.Path(os.path.dirname(os.path.abspath(__file__))).parent
        self.app_path = self.main_path / dir_name_app 
        self.app_data_path = pathlib.Path(os.path.expandvars(r'%LOCALAPPDATA%')) / pathlib.Path(self.company) / pathlib.Path(self.app_name) 
        self.log_path = self.app_data_path / dir_name_log
        self.config_path = self.app_data_path / dir_name_config
        self.config_file =  self.config_path / "last_config.ini"

        """ application specific settings """
        self.config_section = "Application"
        self.resource_path = self.main_path / dir_name_resources 
        self.settings = AppSettings()
        self.modules = {}

    def start_app(self):
        fileutil.create_folder(self.config_path)
        prop_val.load_from_ini_file(self.settings, self.config_file, self.config_section)   
        self.setup_logger()
        self.appwindow = app_gui.AppWindow(self)
        self.appwindow.create_gui(self.resource_path / "app_stylesheet.qss", self.resource_path / "app_icon.ico")
        self.lastWindowClosed.connect(self.quit_app)
        self.do_startup()
        self.activate_module(self.settings.ActiveModuleIndex.value)
        self.appwindow.show()
        self.logger.info("App running...")

    def quit_app(self):
        self.stop_modules()
        self.do_shutdown()
        prop_val.save_to_ini_file(self.settings, self.config_file, self.config_section)   

    def do_startup(self):
        raise NotImplementedError(f"Subclass of '{self.__class__.__name__}' has to implement '{sys._getframe().f_code.co_name}()'")
   
    def do_shutdown(self):
        raise NotImplementedError(f"Subclass of '{self.__class__.__name__}' has to implement '{sys._getframe().f_code.co_name}()'")

    def save_settings(self, module: module_base.ModuleBase):
        if isinstance(module.settings, prop_val.PropStore):
            prop_val.save_to_ini_file(module.settings,self.config_file, module.name)
 
    def load_settings(self, module: module_base.ModuleBase):
        if isinstance(module.settings, prop_val.PropStore):
            if self.settings.LoadLastSettings.value:
                prop_val.load_from_ini_file(module.settings, self.config_file, module.name)
 
    def setup_logger(self, logfile: pathlib.Path = pathlib.Path("latest.log")):   
        """
        Setup Logger if needed and differ between consol and file output
        set up logging to file and choose logger level (from highest level to lowest: Debug, info, warning, error, critical)
        """ 
        if self.settings.Logging.value:
            fileutil.create_folder(self.log_path)
            logfilepath = str(self.log_path / logfile) 
            logging.basicConfig(level=logging.DEBUG,
                                format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                                datefmt='%m-%d %H:%M:%S',
                                filename=logfilepath,
                                filemode='w')
            console = logging.StreamHandler()
            console.setLevel(logging.INFO)
            formatter = logging.Formatter('%(name)-12s %(levelname)-8s %(message)s')
            console.setFormatter(formatter)
            logging.getLogger('').addHandler(console)
        self.logger = logging.getLogger('Application')
        self.logger.setLevel(logging.DEBUG)

    def show_message(self, msg: str, msg_type: app_common.MsgType = app_common.MsgType.Info):
        self.appwindow.show_message(msg, msg_type)

    def add_module(self, new_module: module_base.ModuleBase, name: str):
        new_module.name = name
        self.logger.info(f"Start module: {new_module.name}")
        self.modules[new_module.name] = new_module
        new_module.start()
        self.appwindow.add_screen(new_module)
        self.appwindow.update_menu(self.modules)

    def get_module_count(self) -> int:
        return len(self.modules)     

    def activate_module(self, module_index: int):
        self.settings.ActiveModuleIndex.value = module_index
        self.appwindow.set_active_module_by_index(module_index)   
        
    def stop_modules(self):
        for mod in self.modules.values():
            self.logger.info(f"Stop module: {mod.name}")
            mod.stop()

    def activate_debugger_support_for_this_thread(self):
        """ This function has to be called from each new thread to activate debugger support for it"""
        if self.is_in_debugging_mode:
            import debugpy
            debugpy.debug_this_thread()            
 



""" Defines the configuration values of the module.
    Values in the ProStore are saved and loaded during application startup/shutdown automatically
Copyright Nanosurf AG 2021
License - MIT
"""

import enum
import nanosurf
import nanosurf.lib.datatypes.sci_val as sci_val
import nanosurf.lib.datatypes.prop_val as prop_val

class ScanSettings(prop_val.PropStore):
    """ settings defined here as PropVal are stored persistently in a ini-file
        settings with a '_' as first char are not stored
    """
    def __init__(self):    
        self.image_size = prop_val.PropVal(sci_val.SciVal(10e-6, "m"))
        self.time_per_line = prop_val.PropVal(sci_val.SciVal(0.2, "s"))
        self.points_per_line = prop_val.PropVal(int(256))
        self.channel_id = prop_val.PropVal(int(nanosurf.Spm.ScanChannel.Deflection_or_ZCtrlIn))
        self.show_backward = prop_val.PropVal(bool(True))
        self.show_power_spec = prop_val.PropVal(bool(True))
        self.show_compress_spec = prop_val.PropVal(bool(False))

class ScanResults():
    """ This class saves the worker task result (e.g be read by gui elements or saved to file """
    def __init__(self):
        self.current_scan_line_index = 0

""" Combo boxes show entries with IDs and the names in second list"""
class ChannelD(enum.IntEnum):
    Deflection = 0,
    Topography = 1,




""" This is the screen of the module
Copyright Nanosurf AG 2021
License - MIT
"""

from PySide2 import QtWidgets
import nanosurf as nsf
from app import module_base, app_gui
from modules.demo_module import module, settings

class ResultTableID(nsf.gui.nsf_tables.TableEntryIDs):
    """ identifier id are used in a nsf_table widget"""
    Items = 0
    last_data = 1
    mean_value = 2

ComboPlotStyles = [
    nsf.gui.NSFComboEntry(settings.PlotStyleID.PlotSin,'Sin-Plot'),
    nsf.gui.NSFComboEntry(settings.PlotStyleID.PlotCos,'Cos-Plot'),
]

""" some useful list of allowed prefixes used by nsf_sci_edit widgets"""
allowed_count_units = [nsf.sci_val.up.Prefix.base]
allowed_time_units = [nsf.sci_val.up.Prefix.base, nsf.sci_val.up.Prefix.milli]
allowed_meter_units = [nsf.sci_val.up.Prefix.milli, nsf.sci_val.up.Prefix.micro, nsf.sci_val.up.Prefix.nano]

class DemoScreen(module_base.ModuleScreen):
    def __init__(self):
        super().__init__()

    def do_setup_screen(self, module: module.DemoModule):
        """ create here your gui with all controls and their layout"""
        self.module = module

        # left layout - main controls ------------------------------------------------------------

        self.scival_repetitions = nsf.gui.NSFSciEdit("Repetitions")
        self.scival_repetitions.set_allowed_prefix_ids(allowed_count_units)
        self.scival_repetitions.set_prefix_id(nsf.sci_val.up.Prefix.base)
        self.scival_repetitions.set_precision(0)
        self.scival_repetitions.set_value_min_max(1, 1000)

        self.scival_time_per_rep = nsf.gui.NSFSciEdit("Time per Repetition")
        self.scival_time_per_rep.set_allowed_prefix_ids(allowed_time_units)
        self.scival_time_per_rep.set_prefix_id(nsf.sci_val.up.Prefix.base)
        self.scival_time_per_rep.set_precision(2)
        self.scival_time_per_rep.set_value_min_max(0.01, 10.0)
        
        self.check_emit_ticks = QtWidgets.QCheckBox("Emit Ticks")
        self.button_start_stop = QtWidgets.QPushButton("Start")

        self.layout_left = QtWidgets.QVBoxLayout()
        self.layout_left.addWidget(self.scival_repetitions)
        self.layout_left.addWidget(self.scival_time_per_rep)
        #self.layout_left.addSpacerItem(app_gui.StdVSpacer())
        self.layout_left.addWidget(self.check_emit_ticks)
        self.layout_left.addStretch()
        self.layout_left.addWidget(self.button_start_stop)

        # mid layout - plots and result ---------------------------------------------------------
        self.chart_plot = nsf.gui.NSFChart(logmodex=False)

        self.tableResults = nsf.gui.NSFNameValueTable(ResultTableID)
        self.tableResults.define_entry(ResultTableID.Items,"Items")
        self.tableResults.define_entry(ResultTableID.last_data,"Last Data")
        self.tableResults.define_entry(ResultTableID.mean_value,"Mean Value")

        self.layout_mid  = QtWidgets.QVBoxLayout()
        self.layout_mid.addWidget(self.chart_plot)
        self.layout_mid.addWidget(self.tableResults)

        # right layout - additional user inputs
        self.combo_plot_style = nsf.gui.NSFComboBox(ComboPlotStyles,"Plot Style")
        
        self.layout_right= QtWidgets.QVBoxLayout()
        self.layout_right.addWidget(self.combo_plot_style)
        self.layout_right.addSpacerItem(app_gui.StdVSpacer())

        # set GUI controls
        self.screen_layout = QtWidgets.QHBoxLayout()
        # stretch only plot area and keep controls fix in size
        self.screen_layout.addLayout(self.layout_left, 0)
        self.screen_layout.addLayout(self.layout_mid,  1) 
        self.screen_layout.addLayout(self.layout_right,0)
        self.setLayout(self.screen_layout)

        self.bind_gui_elements()
        self.init_plot()
        self.enter_gui_state_idle()

    def bind_gui_elements(self):
        """ connect all gui widgets to module settings or any other source """
        
        # binding ProVal to widgets ensure that they are alway in sync 
        nsf.gui.connect_to_property(self.combo_plot_style, self.module.settings.plot_func_id)
        nsf.gui.connect_to_property(self.scival_repetitions, self.module.settings.repetitions)
        nsf.gui.connect_to_property(self.scival_time_per_rep, self.module.settings.time_per_repetition)
        nsf.gui.connect_to_property(self.check_emit_ticks, self.module.settings.send_ticks)

        # buttons have to be connected separately
        self.button_start_stop.clicked.connect(self.on_button_start_stop_clicked)    
        
        # listen to signals from the core module to react and update the gui
        self.module.sig_work_start_requested.connect(self.enter_gui_state_wait)
        self.module.sig_work_stop_requested.connect(self.enter_gui_state_wait)
        self.module.sig_work_active.connect(self.enter_gui_state_active)
        self.module.sig_work_done.connect(self.enter_gui_state_idle)
        self.module.sig_new_data_available.connect(self.show_new_data)
        self.module.sig_data_invalid.connect(self.set_data_invalid)

    def init_plot(self):
        self.chart_plot.set_title(f"Measurements")
        self.chart_plot.set_label(nsf.gui.NSFChart.Axis.bottom, "Index")
        self.chart_plot.set_unit(nsf.gui.NSFChart.Axis.bottom, "")
        self.chart_plot.set_label(nsf.gui.NSFChart.Axis.left, "Data")
        self.chart_plot.set_unit(nsf.gui.NSFChart.Axis.left, "Arb")  
        self.chart_plot.plot.setXRange(0, self.module.settings.repetitions.value-1)
        # self.chart_plot.plot.vb.enableAutoRange(y=True)
        # self.chart_plot.plot.vb.setAutoVisible(y=False)
        self.chart_plot.clear_plots()

    def on_button_start_stop_clicked(self):
        if self.module.is_worker_busy():
            self.module.stop_worker()
        else:
            self.module.start_worker()

    def enter_gui_state_wait(self):
        self.set_parameter_widget_enable_state(enabled=False)
        self.start_stop_button_state(wait=True)

    def enter_gui_state_active(self):
        self.set_parameter_widget_enable_state(enabled=False)
        self.start_stop_button_state(wait=False, stop_state=self.module.is_worker_busy())
        self.init_plot()

    def enter_gui_state_idle(self):
        self.set_parameter_widget_enable_state(enabled=True)
        self.start_stop_button_state(wait=False, stop_state=self.module.is_worker_busy())

    def set_parameter_widget_enable_state(self, enabled: bool = True):
        self.scival_time_per_rep.setEnabled(enabled)
        self.scival_repetitions.setEnabled(enabled)
        self.combo_plot_style.setEnabled(enabled)

    def start_stop_button_state(self, wait: bool = False, stop_state: bool = False):
        if wait:
            self.button_start_stop.setEnabled(False)
            self.button_start_stop.setText("Wait...")
        else:
            self.button_start_stop.setEnabled(True)
            self.button_start_stop.setText("Stop" if stop_state else "Start")

    def show_new_data(self):
        self.update_plot()
        self.update_result()

    def set_data_invalid(self):
        self.chart_plot.clear_plots()
        self.tableResults.clear_values()

    def update_result(self):
        res = self.module.get_result()
        self.tableResults.set_value(ResultTableID.Items, res.number_of_data_points, "", precision=0)
        self.tableResults.set_value(ResultTableID.last_data, res.last_data, "")
        self.tableResults.set_value(ResultTableID.mean_value, res.mean_value, "")

    def update_plot(self): 
        current_data = self.module.get_worker_result()
        self.chart_plot.plot_data(y=current_data.value)



"""Package for scripting the Nansurf control software.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import time
from nanosurf.lib.spm import scanhead
import nanosurf
from nanosurf.lib.spm.scanhead import motor_control as mc

class DriveMotorID():
    ALLMOTORS = mc.MotorID(-1, 00, 0)
    ALLAPPROACHMOTORS = mc.MotorID(nanosurf.Spm.SystemMotorID.Approach, 1000, 0.0101)
    APPROACHMOTOR1 =  mc.MotorID(nanosurf.Spm.SystemMotorID.FootA, 1001, 0.0101)
    APPROACHMOTOR2 = mc.MotorID(nanosurf.Spm.SystemMotorID.FootB, 1002, 0.0101)
    APPROACHMOTOR3 = mc.MotorID(nanosurf.Spm.SystemMotorID.FootC, 1003, 0.0101)
    FOCUS = mc.MotorID(nanosurf.Spm.SystemMotorID.Focus, 1004, 0.0056)
    PTE_X = mc.MotorID(nanosurf.Spm.SystemMotorID.PTEX, 1005, 0.0008)
    PTE_Y = mc.MotorID(nanosurf.Spm.SystemMotorID.PTEY, 1006, 0.0008)
    READOUT_X = mc.MotorID(nanosurf.Spm.SystemMotorID.BeamDefX, 1007, 0.0008)
    READOUT_Y = mc.MotorID(nanosurf.Spm.SystemMotorID.BeamDefY, 1008, 0.0008)
    PHOTODIODE_X = mc.MotorID(nanosurf.Spm.SystemMotorID.PhotoDetLateral, 1009, 0.0033)
    PHOTODIODE_Y = mc.MotorID(nanosurf.Spm.SystemMotorID.PhotoDetNormal, 1010, 0.0033)
    LENSEGIMBAL = mc.MotorID(nanosurf.Spm.SystemMotorID.LensGimbal, 1013, 40)
    XAXIS = mc.MotorID(-1, 1011, -1)
    YAXIS = mc.MotorID(-1, 1012, -1) 

class DriveAFMScanhead(scanhead.Scanhead):   
    def __init__(self, spm: nanosurf.Spm = None, *args, **kargs):
        super().__init__(spm, *args, **kargs)

        self.__pd_adjust_tolerance = 0.005

    def do_connect(self) -> bool:
        if self.spm.is_lowlevel_scripting_enabled():
            self.lu_sensor_ctrl = self.spm.lowlevel.SensorControl(self.spm.lowlevel.SensorControl.Instance.SGLE)
            self.motor_control = mc.MotorControl()
            self.connected = self.motor_control.connect(self.spm)
        else:
            self.logger.error("Lowlevel Scripting interface of spm controller is not enabled.")
        return self.connected

    @property
    def pd_adjust_tolerance(self):
        return self.__pd_adjust_tolerance

    @pd_adjust_tolerance.setter
    def pd_adjust_tolerance(self, val):
        """ """
        self.__pd_adjust_tolerance = val

    def get_motor_control(self) -> mc.MotorControl:
        return self.motor_control

    def start_photo_detector_auto_adjustment(self):
        """ Auto zeroing of photo detector. """
        self.lu_sensor_ctrl.detector_adjustment_tolerance.value = self.pd_adjust_tolerance
        self.lu_sensor_ctrl.start_detector_auto_adjustment()
        time.sleep(self.motor_control.sleep_time_after_motor_command)
        
    def is_photo_detector_auto_adjustment_running(self) -> bool:
        """ """
        running = self.lu_sensor_ctrl.detector_auto_adjustment_status.value != 0.0
        return running

    def stop_photo_detector_auto_adjustment(self):
        self.lu_sensor_ctrl.user_abort()
        
    def do_photo_detector_auto_adjustment(self, timeout: float = 60.0) -> bool:
        """Function to start the AutoZeroing and wait until terminated (uses the three other functions above). 
        If the adjustment terminated not successfully after 30 seconds it is stopped and return value is false. 
        """
        self.logger.info("Start Centering Photodiode")
        self.start_photo_detector_auto_adjustment()

        success = True
        start_time = time.time()
        while self.is_photo_detector_auto_adjustment_running():
            if (time.time() - start_time) > timeout:
                self.stop_photo_detector_auto_adjustment()
                self.logger.error("Centering of Photodiode Failed")
                success = False
                break
            time.sleep(0.5)
        self.stop_photo_detector_auto_adjustment()
        self.motor_control.stop_all_motors()
        if success:
            self.logger.info("Centering of Photodiode Succeeded")
        return success

    def do_move_lasers_to_zero(self) -> bool:
        """ Move both lasers to zero position
        """
        motors = DriveMotorID()
        laser_motors = [motors.PTE_X, motors.PTE_Y, motors.READOUT_X, motors.READOUT_Y]
        done = self.motor_control.do_move_absolute(laser_motors , [0.0 for i in laser_motors])

        if done:
            self.logger.debug("Lasers are moved to zero position without referencing")
        else: 
            self.logger.error("Error or timeout while moving lasers to zero position")   
        return done

    def do_move_all_optics_to_zero(self) -> bool:
        """ Centers all the optical adjustment motors: 
            Lasers and photodiode axis are centered.
            The lensegimbal and the focus are set to the position for air
        """
        motors = DriveMotorID()
        optics_motors = [
            motors.PTE_X, motors.PTE_Y, motors.READOUT_X, motors.READOUT_Y, 
            motors.PHOTODIODE_X, motors.PHOTODIODE_Y, 
            motors.LENSEGIMBAL, motors.FOCUS
        ]
        done = self.motor_control.do_move_absolute(optics_motors, [0.0 for i in optics_motors])

        if done:
            self.logger.debug("All optics motors are moved to zero position without referencing")
        else: 
            self.logger.error("Error or timeout while moving motors to zero position")   
        return done
    
    def do_move_all_optics_to_initial_position(self) -> bool:
        """ Centers all the optical adjustment motors: 
            Lasers and photodiode axis are centered.
            The lensegimbal and the focus are set to the position for air
        """
        done = True
        motors = DriveMotorID()
        motors_for_centering = [
            motors.PTE_X, motors.PTE_Y, motors.READOUT_X, motors.READOUT_Y, 
            motors.PHOTODIODE_X, motors.PHOTODIODE_Y
        ]
        motors_to_position = [motors.LENSEGIMBAL, motors.FOCUS]
        move_position_list = [20, 0.0028] # 20degrees should be vertically aligned
        if done:
            done &= self.motor_control.do_motor_reference_and_center(motors_for_centering)
        if done:
            done &= self.motor_control.do_motor_referencing(motors_to_position)
        if done:
            done &= self.motor_control.do_move_relative(motors_to_position, move_position_list)
        if done:
            self.motor_control.set_motor_zero_position(motors_for_centering + motors_to_position)

        if done:
            self.logger.debug("All optics motors are initial positionreferenced, centered and zeroed")    
        else:
            self.logger.error("Error or timeout while moving motors to initial position")    
        return done

    def do_move_all_feet_motors_to_full_range_then_center(self) -> bool:
        """ Move feet motors to full extension and full retraction and then to mid position 
        """
        done = True
        all_approach_motor_id = DriveMotorID().ALLAPPROACHMOTORS
        if done:
            done &= self.motor_control.do_motor_referencing(all_approach_motor_id)
        if done: 
            done &= self.motor_control.do_move_relative(all_approach_motor_id, 1) # We move full range (just large enough number and wait until motor stopped)
        if done:
            mid_range = self.motor_control.get_motor_full_range(all_approach_motor_id) / 2.0
            done &= self.motor_control.do_move_relative(all_approach_motor_id, -mid_range)
        if done:
            self.logger.debug("All approach feeds are moved to center position")    
        else:
            self.logger.error("Error or timeout while moving approach motors to center position")    
        return done

    def do_reference_and_move_back_lasers(self) -> bool:
        """ Reference and Move Back all laser motors (they should overlay optically, during the scanhead calibration, the lasers should be both
        in the optical center """
        done = True
        motors = DriveMotorID()
        motors_for_centering = [motors.PTE_X, motors.PTE_Y, motors.READOUT_X, motors.READOUT_Y]
        done &= self.motor_control.do_reference_and_move_back(motors_for_centering)
        return done


"""Package for scripting the Nansurf control software.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import nanosurf.lib.spm.lowlevel.ctrlunits as ctrlunits

class _CtrlUnitDAC(ctrlunits._CtrlUnit):
    def __init__(self, spm, lu_inst_name):
        super().__init__(spm)
        self._lu_inst_name = lu_inst_name

    @property
    def inst_name(self):
        return self._lu_inst_name
    @property
    def dc(self):
        return self._lu.current_output_value.value

    @dc.setter
    def dc(self, val):
        self._lu.static_value.value = val
        self._lu.input.value = self._static_input_channel_index

    @property
    def dc_norm(self):
        return self._lu.current_output_value.value / self.max

    @dc_norm.setter
    def dc_norm(self, val):
        self.dc = val * self.max

    @property
    def dc_v(self):
        return self.dc_norm * self._voltage_range

    @dc_v.setter
    def dc_v(self, val):
        self.dc_norm = val / self._voltage_range

    @property
    def unit(self):
        return self._lu.current_output_value.unit

    @property
    def max(self):
        return self._lu.current_output_value.value_max

    @property
    def source(self):
        return self.__source

    @source.setter
    def source(self, val):
        import nanosurf.lib.spm.lowlevel.ctrlunits.analyzer as analyzer
        import nanosurf.lib.spm.lowlevel.ctrlunits.channelmux as channelmux

        self.__source = val
        if isinstance(self.__source, analyzer.CtrlUnitsSineWaveGenerator):
            self.__source.target = self
            self._lu.input.value = self._ac_input_channel_index[self.__source.inst_name]
            self._lu.calib_sig_source_dir.value = self._spm.lowlevel.AnalogHiResOut.CalibSigSourceDir.FromOutputToInput
        elif isinstance(self.__source, channelmux._CtrlUnitChannelMux):
            self.__source.channelmap = self._channelmap
            self.__source.target = self
            if self.__source.channel != "":
                self.__source.channel =  self.__source.channel  # reset channel to saved value
            self._lu.calib_sig_source_dir.value = self._spm.lowlevel.AnalogHiResOut.CalibSigSourceDir.FromInputToOutput
        else:
            self._lu.input.value = self._static_input_channel_index
            self._lu.calib_sig_source_dir.value = self._spm.lowlevel.AnalogHiResOut.CalibSigSourceDir.FromOutputToInput

    @property
    def inverted(self):
        return self._lu.calib_polarity.value == self._spm.lowlevel.AnalogHiResIn.CalibPolarity.Negative

    @inverted.setter
    def inverted(self, val):
        if val == True:
            self._lu.calib_polarity.value = self._spm.lowlevel.AnalogHiResIn.CalibPolarity.Negative
        else:
            self._lu.calib_polarity.value = self._spm.lowlevel.AnalogHiResIn.CalibPolarity.Positive

    @property
    def calib_gain(self):
        return self._lu.calib_gain.value

    @calib_gain.setter
    def calib_gain(self, val):
        self._lu.calib_gain.value = val

    @property
    def calib_offset(self):
        return self._lu.calib_offset.value

    @calib_offset.setter
    def calib_offset(self, val):
        self._lu.calib_offset.value = val

class CtrlUnitDACHires(_CtrlUnitDAC):
    def __init__(self, spm, lu_inst_name):
        super().__init__(spm, lu_inst_name)
        self._lu = spm.lowlevel.AnalogHiResOut(spm.lowlevel.AnalogHiResOut.Instance[lu_inst_name])
        self._voltage_range = 10.0
        self._channelmap = self._spm.lowlevel.AnalogHiResOut.InputChannels
        self._static_input_channel_index = spm.lowlevel.AnalogHiResOut.InputChannels.Static
        self._ac_input_channel_index = {
            "INST2":spm.lowlevel.AnalogHiResOut.InputChannels.GenTest_Dynamic
        }

class CtrlUnitDACFast(_CtrlUnitDAC):
    def __init__(self, spm, lu_inst_name):
        super().__init__(spm, lu_inst_name)
        self._lu = spm.lowlevel.AnalogFastOut(spm.lowlevel.AnalogFastOut.Instance[lu_inst_name])
        self._voltage_range = 1.0
        self._channelmap = self._spm.lowlevel.AnalogFastOut.InputChannels
        self._static_input_channel_index = spm.lowlevel.AnalogFastOut.InputChannels.Static
        self._ac_input_channel_index = {
            "INST1":spm.lowlevel.AnalogFastOut.InputChannels.Analyzer1_Reference,
            "INST2":spm.lowlevel.AnalogFastOut.InputChannels.Analyzer2_Reference,
        }



"""Package for scripting the Nanosurf control software.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import os
import pathlib
from nanosurf._version import __version__

# make all library modules accessible from this level
from nanosurf.lib import datatypes, devices, gui, math, spm,  util

# direct access to most used datatypes
from nanosurf.lib.datatypes import sci_val, nsf_thread
from nanosurf.lib.datatypes.sci_stream import SciStream
from nanosurf.lib.datatypes.sci_val import SciVal
from nanosurf.lib.datatypes.sci_channel import SciChannel
from nanosurf.lib.datatypes.prop_val import PropStore, PropVal

# other direct access to common sub packages
from nanosurf.lib.math import sci_math 
from nanosurf.lib.spm import studio
from nanosurf.lib.spm.com_proxy import Spm

class Studio(studio.StudioScriptSession):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

class SPM(Spm):
    def __init__(self, *args, **kwargs):
        super().__init__("SPM.Application", *args, **kwargs)

class USPM(Spm):
    def __init__(self, *args, **kwargs):
        super().__init__("USPM.Application", *args, **kwargs)

class CX(Spm):
    def __init__(self, *args, **kwargs):
        super().__init__("CX.Application", *args, **kwargs)

class C3000(Spm):
    def __init__(self, *args, **kwargs):
        super().__init__("C3000.Application", *args, **kwargs)

class Naio(Spm):
    def __init__(self, *args, **kwargs):
        super().__init__("Naio.Application", *args, **kwargs)

class CoreAFM(Spm):
    def __init__(self, *args, **kwargs):
        super().__init__("CoreAFM.Application", *args, **kwargs)

class Easyscan2(Spm):
    def __init__(self, *args, **kwargs):
        super().__init__("Easyscan2.Application", *args, **kwargs)

class MobileS(Spm):
    def __init__(self, *args, **kwargs):
        super().__init__("MobileS.Application", *args, **kwargs)

class SPM_S(Spm):
    def __init__(self, *args, **kwargs):
        super().__init__("SPM_S.Application", *args, **kwargs)

def package_path() -> pathlib.Path:
    return pathlib.Path(os.path.dirname(os.path.abspath(__file__)))

def doc_path() -> pathlib.Path:
    return package_path() / "doc"

def lib_path() -> pathlib.Path:
    return package_path() / "lib"

def app_path() -> pathlib.Path:
    return package_path() / "app"

def help(): 
    print("Nanosurf python script package:")
    print("-------------------------------")
    print(f"Installed version: {__version__}")
    print("\nThe library documentation is here:")
    print(doc_path())
    for doc in os.listdir(doc_path()):
        print(f"  {doc}")
    print("\nThe demos and apps are stored here:")
    print(app_path())
    for demo in os.listdir(app_path()):
        print(f"  {demo}")





"""Mock for application.SPMCtrlManager.LogicalUnit object.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""


class SPMCtrlManager():
    def __init__(self):
        self.LogicalUnit = LogicalUnit()
        self.DataBuffer = None


class LogicalUnit():
    """Mock for application.SPMCtrlManager.LogicalUnit object."""
    def __init__(self):
        self._value_store = [[[
            0.0 for attr_id in range(50)]
                for inst_id in range(20)]
                for type_id in range(100)
        ]

    def Attribute(self, type_id, instance_id, attr_id, value=None):
        if value is None:
            value = self._value_store[type_id][instance_id][attr_id]
            print(
                "GetAttr: Type ID: " + str(type_id) +
                " Inst No: " + str(instance_id) +
                " Attr ID: " + str(attr_id) +
                " Value: " + str(value)
            )
            return value
        else:
            self._value_store[type_id][instance_id][attr_id] = value
            print(
                "SetAttr: Type ID: " + str(type_id) +
                " Inst No: " + str(instance_id) +
                " Attr ID: " + str(attr_id) +
                " Value: " + str(value)
            )

    def AttributeMax(self, type_id, instance_id, attr_id):
        return 10.0

    def AttributeMin(self, type_id, instance_id, attr_id):
        return -10.0

    def AttributeUnit(self, type_id, instance_id, attr_id):
        return "m"

    def AttributeVectorDouble(self, type_id, instance_id, attr_id, value=None):
        if value is None:
            value = self._value_store[type_id][instance_id][attr_id]
            print(
                "GetAttr: Type ID: " + str(type_id) +
                " Inst No: " + str(instance_id) +
                " Attr ID: " + str(attr_id) +
                " Value: " + str(value)
            )
        else:
            self._value_store[type_id][instance_id][attr_id] = value
            print(
                "SetAttr: Type ID: " + str(type_id) +
                " Inst No: " + str(instance_id) +
                " Attr ID: " + str(attr_id) +
                " Value: " + str(value)
            )
        return value

    def AttributeVectorByIndexDouble(self, type_id, instance_id, attr_id, index, value=None):
        if value is None:
            value = self._value_store[type_id][instance_id][attr_id][index]
            print(
                "GetItem: Type ID: " + str(type_id) +
                " Inst No: " + str(instance_id) +
                " Attr ID: " + str(attr_id) +
                " Index: " + str(index) +
                " Value: " + str(float(value))
            )
        else:
            self._value_store[type_id][instance_id][attr_id][index] = value
            print(
                "SetItem: Type ID: " + str(type_id) +
                " Inst No: " + str(instance_id) +
                " Attr ID: " + str(attr_id) +
                " Index: " + str(index) +
                " Value: " + str(float(value)))
        return value

    def Trigger(self, type_id, instance_id, trigger_id):
        print(
            "Trig: Type_id: " + str(type_id) +
            " Inst No: " + str(instance_id) +
            " Trigger ID: " + str(trigger_id))

    def IsBusy(self, type_id, instance_id, busy_id):
        print(
            "IsBusy: Type_id: " + str(type_id) +
            " Inst No: " + str(instance_id) +
            " Busy ID: " + str(busy_id))
        return False

if __name__ == '__main__':
    print("\nTest LUI")
    lui = LogicalUnit()
    lui.Trigger(3, 2, 1)
    assert (not lui.IsBusy(3, 2, 5))

    test_val = 0.4
    lui.Attribute(1, 2, 3, test_val)
    assert lui.Attribute(1, 2, 3) == test_val

    test_val = [0.1, 0.2, 0.3]
    lui.AttributeVectorDouble(1, 3, 3, test_val)
    assert lui.AttributeVectorDouble(1, 3, 3) == test_val
    assert lui.AttributeVectorByIndexDouble(1, 3, 3, 0) == test_val[0]

    test_val = 2.0
    lui.AttributeVectorByIndexDouble(1, 3, 3, 1, test_val)
    assert lui.AttributeVectorByIndexDouble(1, 3, 3, 1) == test_val
    del(lui)



"""scientific value library formats and parse nice looking number strings into float numbers
    It handles units and exponent characters e.g. A string "2.45223kg" result in a float number 2452.23 and unit 'g'
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import math
import sys

import nanosurf.lib.datatypes.sci_val.unit_prefix as up
import nanosurf.lib.datatypes.sci_val.convert as convert
import nanosurf.lib.spm.studio.wrapper as studio_wrapper

class SciVal:
    def __init__(self, value = 0.0, unit_str: str = ""):
        if isinstance(value, SciVal):
            self._value = value._value
            self._value_default = value._value_default
            self._value_min = value._value_min
            self._value_max = value._value_max
            self._unit = value._unit
            self._prefix_id = up.Prefix.auto_
            self._allowed_unit_prefixes = value._allowed_unit_prefixes
            self._is_real = value._is_real
        elif isinstance(value, studio_wrapper.CmdTreeProp):
            self._value = value.value
            self._value_default = value.value
            self._value_min = -sys.float_info.max
            self._value_max = sys.float_info.max
            self._unit = value.unit
            self._prefix_id = up.Prefix.auto_
            self._allowed_unit_prefixes = up.default_prefix_range
            self._is_real = True
        else:
            self._value = value
            self._value_default = value
            self._value_min = -sys.float_info.max
            self._value_max = sys.float_info.max
            self._unit = unit_str
            self._prefix_id = up.Prefix.auto_
            self._allowed_unit_prefixes = up.default_prefix_range
            self._is_real = True
          
    def set_prefix_id(self, prefix_id: up.Prefix):
        if (self.is_prefix_enabled(prefix_id) or (prefix_id == up.Prefix.auto_)):
            self._prefix_id = prefix_id

    def prefix_id(self) -> up.Prefix:
        return self._prefix_id    
    
    def prefix_id_to_str(self, prefix_id: up.Prefix) -> str:
        return up.prefix_id_to_string(prefix_id)

    def _change_value_by_prefix_id(self, new_prefix_id: up.Prefix, old_prefix_id: up.Prefix):
        if (new_prefix_id != self._prefix_id and self.can_have_prefix() and self.is_prefix_enabled(new_prefix_id)):
            exp_old_val = up.prefix_id_to_absolute_value(old_prefix_id)
            exp_new_val = up.prefix_id_to_absolute_value(new_prefix_id)
            self.set_value(self._value / exp_old_val * exp_new_val)

    def is_real(self) -> bool:
        return self._is_real

    def to_string(self, precision : int = 3):
        return convert.to_string(self._value, self._unit, precision, self._prefix_id, self._allowed_unit_prefixes)

    def to_string_formatted(self, prefix_id: up.Prefix, precision: int = 3):
        return convert.to_string(self._value, self._unit, precision, prefix_id, up.default_prefix_range)

    def allowed_prefixes(self) -> list[up.Prefix]:
        return self._allowed_unit_prefixes

    def is_prefix_enabled(self, prefix_id: up.Prefix) -> bool:
        return prefix_id in self._allowed_unit_prefixes
        
    def can_have_prefix(self) -> bool:
        no_prefix = (self._unit=="") or (self._unit=="%") or (self._unit=="°") or (self._unit=="°C")
        return not(no_prefix)

    def unit(self) -> str:
        return self._unit

    def value(self) -> float:
        return self._value

    def value_min(self) -> float:
        return self._value_min

    def value_max(self) -> float:
        return self._value_max

    def value_default(self) -> float:
        return self._value_default

    def set_unit(self, unit_str: str):
        self._unit = unit_str

    def set_value(self, val: float):
        self._value = min(max(self._value_min, val), self._value_max)
        self._value = self._round_if_integral(self._value)
        # self._value = self._make_odd_if_step_odd1(self._value)
        
    def set_value_min(self, val: float):
        self._value_min = self._round_if_integral(val)
        self.set_value(self._value)

    def set_value_max(self, val: float):
        self._value_max = self._round_if_integral(val)
        self.set_value(self._value)

    def set_value_default(self, val: float):
        self._value_default = self._round_if_integral(val)

    def set_is_real(self, is_real:bool):
        self._is_real = is_real
        
    def set_allowed_prefixes(self, allowed_prefix_ids: list[up.Prefix]):
        self._allowed_unit_prefixes = allowed_prefix_ids

    def from_string(self, value_str) -> bool:
        value_changed = False
        res = convert.to_value(value_str, self._unit, self._prefix_id) 
        if res.success:
            value_changed = (self._value != res.value)
            self.set_value(res.value)
            if self._prefix_id == up.Prefix.auto_:
                self._prefix_id = res.prefix_id
        return value_changed

    def reset_to_default_value(self):
        self.set_value(self._value_default)
        self._prefix_id = up.Prefix.auto_

    # local methods ---------------------------------------------------

    def _create_result_object(self, new_value: float, other):
        if isinstance(other, SciVal):
            res = SciVal(self)
            res.set_value(new_value)
            return res
        return new_value

    def _get_value_from(self, other):
        if isinstance(other, SciVal):
            return other.value()
        return other

    def _round_if_integral(self, val:float) -> float:
        if self._is_real:
            return val
        return float(math.floor(val+0.5))

    # overwrite magic conversion and math functions ----------------------------------

    def __int__(self):
        return int(self._value)

    def __float__(self):
        return self._value

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string()

    def __add__(self, other):
        return self._create_result_object(self._value + self._get_value_from(other), other)   
              
    def __sub__(self, other):
        return self._create_result_object(self._value - self._get_value_from(other), other)   
              
    def __mul__(self, other):
        return self._create_result_object(self._value * self._get_value_from(other), other)   
              
    def __truediv__(self, other):
        return self._create_result_object(self._value / self._get_value_from(other), other)   

    def __floordiv__(self, other):
        return self._create_result_object(self._value // self._get_value_from(other), other)   

    def __mod__(self, other):
        return self._create_result_object(self._value % self._get_value_from(other), other)   
              
    def __pow__(self, other):
        return self._create_result_object(math.pow(self._value / self._get_value_from(other)), other)   

    def __lt__(self, other):
        return self._value < self._get_value_from(other)
              
    def __gt__(self, other):
        return self._value > self._get_value_from(other)
              
    def __le__(self, other):
        return self._value <= self._get_value_from(other)
              
    def __ge__(self, other):
        return self._value >= self._get_value_from(other)
              
    def __eq__(self, other):
        return self._value == self._get_value_from(other)
              
    def __ne__(self, other):
        return self._value != self._get_value_from(other)
              
    def __iadd__(self, other):
        self._value += self._get_value_from(other)
              
    def __isub__(self, other):
        self._value -= self._get_value_from(other)
              
    def __imul__(self, other):
        self._value *= self._get_value_from(other)
              
    def __idiv__(self, other):
        self._value /= self._get_value_from(other)
              
    def __imod__(self, other):
        self._value %= self._get_value_from(other)
              
    def __neg__(self):
        return self._create_result_object(-self._value)   
              
    def __pos__(self):
        return self._create_result_object(+self._value)   
              
def from_str(val_str: str)-> SciVal:
    res = convert.to_value(val_str, current_unit="", use_current_unit=False)
    if res.success:
        return SciVal(value=res.value, unit_str=res.unit)
    return SciVal()




"""Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""


import typing
import numpy as np

class SciChannel:
    def __init__(self, 
        copy_from: typing.Union[list, np.ndarray, 'SciChannel', None] = None, 
        array_length: int = 0, 
        unit :str = ""):

        self.value = np.zeros(array_length)
        self.unit = unit

        if copy_from is not None:
            if isinstance(copy_from, SciChannel):
                self.value = np.copy(copy_from.value)
                self.unit = copy_from.unit
            else:
                self.value = np.copy(copy_from)
                self.unit = unit





""" application wide info box
Copyright Nanosurf AG 2021
License - MIT
"""
from PySide2 import QtWidgets
from nanosurf.lib.gui import nsf_colors

class InfoBox(QtWidgets.QWidget):
    def __init__(self, hidden : bool = True):
        super().__init__()
        self.set_background_color(nsf_colors.NSFColorHexStr.Orange)
        self.set_text_color("000000")
        self.message = QtWidgets.QLabel()
        self.message.setFixedHeight(30)
        self.closeButton = QtWidgets.QPushButton("X")
        self.closeButton.setSizePolicy(QtWidgets.QSizePolicy.Fixed,QtWidgets.QSizePolicy.Fixed)        
        self.box_layout = QtWidgets.QHBoxLayout()
        self.box_layout.addWidget(self.message)
        self.box_layout.addWidget(self.closeButton)
        self.setLayout(self.box_layout)
        self.setMaximumHeight(60)
        self.setHidden(hidden)
        self.closeButton.clicked.connect(self._on_close_button_clicked)
    
    def set_message(self, msg: str):
        self.message.setText(f"    {msg}")
        self.show_box(True)

    def show_box(self, show: bool):
        self.setHidden(not(show))

    def set_background_color(self, rgb_color_str: str):
        self.setStyleSheet(f"background-color:#{rgb_color_str};")

    def set_text_color(self, rgb_color_str: str):
        pass
        # self.setStyleSheet(f"color:#{rgb_color_str};")

    def _on_close_button_clicked(self):
        self.setHidden(True)

class NSFInfoBox(InfoBox):
    pass



"""scientific value library formats and parse nice looking number strings into float numbers
    It handles units and exponent characters e.g. A string "2.45223kg" result in a float number 2452.23 and unit 'g'
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import enum

class Prefix(enum.IntEnum):
    yotta =  0,
    zetta =  1,
    exa   =  2,
    peta  =  3,
    tera  =  4,
    giga  =  5,
    mega  =  6,
    kilo  =  7,
    base  =  8,
    milli =  9,
    micro = 10,
    nano  = 11,
    pico  = 12,
    femto = 13,
    atto  = 14,
    zepto = 15,
    yocto = 16,
    auto_ = 17,

  # Full range of prefixes, used as default for sci_val and sci_val_formatter
default_prefix_range = [
    Prefix.yotta,
    Prefix.zetta,
    Prefix.exa,
    Prefix.peta,
    Prefix.tera,
    Prefix.giga,
    Prefix.mega,
    Prefix.kilo,
    Prefix.base,
    Prefix.milli,
    Prefix.micro,
    Prefix.nano,
    Prefix.pico,
    Prefix.femto,
    Prefix.atto,
    Prefix.zepto,
    Prefix.yocto
]


prefix_max_exp_value = +24
prefix_min_exp_value = -24

_prefix_symbols = [
    ["Y","Yotta", 24, 1000000000000000000000000.0],
    ["Z","Zetta", 21,    1000000000000000000000.0],
    ["E", "Exa",  18,       1000000000000000000.0],
    ["P", "Peta", 15,          1000000000000000.0],
    ["T", "Tera", 12,             1000000000000.0],
    ["G", "Giga",  9,                1000000000.0],
    ["M", "Mega",  6,                   1000000.0],
    ["k", "kilo",  3,                      1000.0],
    [ "", "",      0,                         1.0],
    ["m", "milli",-3,                           0.001],
    ["µ", "micro",-6,                           0.000001],
    ["n", "nano", -9,                           0.000000001],
    ["p", "pico", -12,                          0.000000000001],
    ["f", "femto",-15,                          0.00000000000001],
    ["a", "atta", -18,                          0.00000000000000001],
    ["z", "zepto",-21,                          0.00000000000000000001],
    ["y", "yocto",-24,                          0.00000000000000000000001],
    ["",  "auto",   0,                        1.0]
]

class _PrefixSympolID(enum.IntEnum):
    Symbol = 0,
    Name = 1,
    ExpValue = 2,
    AbsValue = 3

def string_prefix_to_prefix_id(prefix_str: str) -> Prefix:
    if prefix_str == "u":
        return Prefix.micro

    for index,p in enumerate(_prefix_symbols):
        if p[_PrefixSympolID.Symbol] == prefix_str:
            prefix_index = index   
            return default_prefix_range[prefix_index]
    return Prefix.base

def is_prefix_str_valid(prefix_str:str) -> bool:
    if prefix_str == "u":
        return True
        
    for p in _prefix_symbols:
        if p[_PrefixSympolID.Symbol] == prefix_str:
            return True   
    return False

def prefix_id_to_string(prefix_id: Prefix):
    return _prefix_symbols[int(prefix_id)][_PrefixSympolID.Symbol] 

def prefix_value_to_string(exponent_value):
    expstr = _prefix_symbols[Prefix.auto_][_PrefixSympolID.Symbol] 
    for p in _prefix_symbols:
        if p[_PrefixSympolID.ExpValue] == exponent_value:
            return p[_PrefixSympolID.Symbol]   
    return expstr

def prefix_id_to_exponent_value(prefix_id:Prefix):
    return _prefix_symbols[prefix_id][_PrefixSympolID.ExpValue] 

def prefix_id_to_absolute_value(prefix_id:Prefix):
    return _prefix_symbols[prefix_id][_PrefixSympolID.AbsValue] 

def prefix_value_to_prefix_id(exp_value):
    prefix_id = Prefix.base
    for index,p in enumerate(_prefix_symbols):
        if p[_PrefixSympolID.ExpValue] == exp_value:
            prefix_id = index   
    return prefix_id


def next_smaller_prefix_id(prefix_id:Prefix):
    if (prefix_id == Prefix.auto_) or (prefix_id == Prefix.yocto):
        return prefix_id
    else:
        for index,p in enumerate(default_prefix_range):
            if p == prefix_id:
                return default_prefix_range[index + 1]
    return  prefix_id           
    




""" The long lasting worker thread as demo - just wait some time and create data periodically
Copyright Nanosurf AG 2021
License - MIT
"""
import time
import math
from PySide2.QtCore import Signal
from nanosurf.lib.datatypes import nsf_thread
from app import app_common, module_base
from modules.demo_module import settings

class MyWorkerData():
    def __init__(self):
        self.value = []
        self.last_index = -1

class MyWorker(nsf_thread.NSFBackgroundWorker):
    """ This class implements the long lasting activity in the background to not freeze the gui """

    sig_tick = Signal() # is emitted if par_send_tick is True
    sig_new_data = Signal() # is emitted each time new data are available . Result is read by self.get_result()

    """ parameter for the background work"""
    par_repetition = 10
    par_time_per_repetition = 2.0
    par_send_ticks = True
    par_time_between_ticks = 0.1
    par_plot_func_id = settings.PlotStyleID.PlotSin

    _sig_message = Signal(str, int)

    def __init__(self, my_module: module_base.ModuleBase):
        self.module = my_module
        self.resulting_data = MyWorkerData()
        super().__init__()
        self._sig_message.connect(self.module.app.show_message)

    def do_work(self):
        """ This is the working function for the long task"""
        # clear data
        self.resulting_data = MyWorkerData()

        self.count = 0
        while (self.count < self.par_repetition) and not self.is_stop_request_pending():
            self.count += 1

            if self.par_send_ticks:
                self.tick_time = 0.0
                while self.tick_time < self.par_time_per_repetition and not self.is_stop_request_pending():
                    self.tick_time += self.par_time_between_ticks
                    time.sleep(self.par_time_between_ticks)
                    self.sig_tick.emit()
                    self.send_message(f"Tick: {self.tick_time:.2f}s")
            else:
                time.sleep(self.par_time_per_repetition)        
            
            if self.par_plot_func_id.value == settings.PlotStyleID.PlotSin:
                self.resulting_data.value.append(math.sin(self.count / self.par_repetition * math.pi))
            elif self.par_plot_func_id.value == settings.PlotStyleID.PlotCos:
                self.resulting_data.value.append(math.cos(self.count / self.par_repetition * math.pi))
            else:
                self.send_message("Unknown plot style", app_common.MsgType.Error)

            self.resulting_data.last_index += 1
            self.sig_new_data.emit()

    def get_result(self) -> MyWorkerData:
        return self.resulting_data

    def send_message(self, msg:str, msg_type: app_common.MsgType = app_common.MsgType.Info):
        self._sig_message.emit(msg, msg_type)
        self.logger.info(msg)   
 


# ///
# /// Line Edit for advanced number formatting
# ///
# /// Copyright (C) Nanosurf AG - All Rights Reserved (2021)
# /// Unauthorized copying of this file, via any medium is strictly prohibited
# /// https://www.nanosurf.com
# ///


from enum import IntEnum
import typing
from PySide2 import QtGui, QtWidgets
from PySide2.QtCore import Qt, Signal

class NSFComboEntry():
    id: int
    name : str

    def __init__(self, id: typing.Union[int, IntEnum], name: str) -> None:
        self.id = int(id)
        self.name = name


class _ComboboxQt(QtWidgets.QComboBox):

    value_changed_event = Signal(float)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._block_value_change_signal = False
        self.currentIndexChanged.connect(self._index_changed)

    # public interface ------------------------------------------

    def set_index(self, new_index: int, notify: bool = True):
        do_notify = notify and (new_index != self.currentIndex())
        self.setCurrentIndex(new_index)
        if do_notify:
            self._index_changed()

    def index(self) -> int:
        return int(self.currentIndex())  

    # internal functions --------------------------------------------    

    def _index_changed(self):
        if not(self._block_value_change_signal):
            self.value_changed_event.emit(self.index())    


class NSFComboBox(QtWidgets.QWidget):
    """ Custom Qt Widget to show text in a edit with a descriptive label. """
    value_changed_event = Signal(str)

    def __init__(self, entries: list[NSFComboEntry] = [], label_str = "", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._setup_widgets(label_str)
        self.comp_entries = entries
        self.define_entries(self.comp_entries)
        self._combo.value_changed_event.connect(self._on_value_changed)

    def define_entries(self, entries: list[NSFComboEntry]):
        self.comp_entries = entries
        self._combo.clear()
        names = [i.name for i in self.comp_entries]
        self._combo.addItems(names)
 
    def set_value(self, id: int, notify: bool = True):
        self._combo.set_index(self._get_index_from_id(id))
        if notify:
            self._on_value_changed()        
       
    def set_label(self, label: str):
       self._label.setText(label) 

    def label(self) -> str:
        return self._label.text()

    def value(self) -> int:
        return self._get_id_from_index(self._combo.index())  

    def current_entry_name(self) -> str:
        return self._combo.currentText()

    def entry_name(self, id : int) -> str:
        return self.comp_entries[self._get_index_from_id(id)].name


    # internal ----------------------------------------------------
    
    def _setup_widgets(self, label_str: str):
        layout = QtWidgets.QVBoxLayout()
        self._label = QtWidgets.QLabel()
        self._label.setText(label_str)
        self._label.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        layout.addWidget(self._label, alignment=Qt.AlignBottom)
        self._combo = _ComboboxQt()
        self._combo.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        layout.addWidget(self._combo)
        self.setLayout(layout)

    def _on_value_changed(self):
        self.value_changed_event.emit(self.value())

    def _get_index_from_id(self, id: int) -> int:
        found_index = -1
        index = 0
        for entry in self.comp_entries:
            if entry.id == id:
                found_index = index
                break
            index += 1
        return found_index

    def _get_id_from_index(self, index: int) -> int:
        return self.comp_entries[index].id





"""Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""


import nanosurf.lib.devices.i2c.bus_master as i2c 

class Chip_PCA9534(i2c.I2CChip):
    """ 8bit GPIO chip"""

    def __init__(self, bus_addr: int, **kwargs):
        super().__init__(bus_addr, i2c.I2COffsetMode.NoOffset, **kwargs)

    @property
    def reg_config(self):
        self.write_byte(0x03)
        return self.read_byte()

    @reg_config.setter
    def reg_config(self, config_mask: int):
        self.write_bytes([0x03, config_mask])

    @property
    def reg_output(self) -> int:
        self.write_byte(0x01)
        return self.read_byte()

    @reg_output.setter
    def reg_output(self, output_bits: int):
        self.write_bytes([0x01, output_bits])

    @property
    def reg_polarity(self):
        self.write_byte(0x02)
        return self.read_byte()

    @reg_polarity.setter
    def reg_polarity(self, bitmask: int):
        self.write_bytes([0x02, bitmask])

    @property
    def reg_input(self) -> int:
        self.write_byte(0x00)
        return self.read_byte()




# studio_wrapper.py

from enum import Enum
from typing import Any
import nanosurf.lib.spm.studio.wrapper as wrap

g_cmd_tree_hash = 'daa181bb17cb3aec4be91c616269c6e0'
g_compiler_version = '1.0'

class RootUtil(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.util'

    def filter_string_array_begin(self, *args) -> Any:
        return self._context.call('root.util.filter_string_array_begin', *args)

    def list_table_tables(self, *args) -> Any:
        return self._context.call('root.util.list_table_tables', *args)

    def list_table_all(self, *args) -> Any:
        return self._context.call('root.util.list_table_all', *args)

    def to_string(self, *args) -> Any:
        return self._context.call('root.util.to_string', *args)

    def list_table_functions(self, *args) -> Any:
        return self._context.call('root.util.list_table_functions', *args)

    def table_append(self, *args) -> Any:
        return self._context.call('root.util.table_append', *args)

    def list_table_vars(self, *args) -> Any:
        return self._context.call('root.util.list_table_vars', *args)

    def make_property(self, *args) -> Any:
        return self._context.call('root.util.make_property', *args)

    def array_concat(self, *args) -> Any:
        return self._context.call('root.util.array_concat', *args)

    def list_table_elements(self, *args) -> Any:
        return self._context.call('root.util.list_table_elements', *args)

    def deep_copy(self, *args) -> Any:
        return self._context.call('root.util.deep_copy', *args)


class RootCoreCore_options_store(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.core.core_options_store'


class RootCoreVshi(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.core.vshi'


class RootCoreScript_server(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.core.script_server'


class RootCoreHw_modules(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.core.hw_modules'


class RootCoreStorage(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.core.storage'

    def close_file(self, *args) -> Any:
        return self._context.call('root.core.storage.close_file', *args)

    def open_file(self, *args) -> Any:
        return self._context.call('root.core.storage.open_file', *args)

    def is_file_open(self, *args) -> Any:
        return self._context.call('root.core.storage.is_file_open', *args)


class RootCoreSpm_controller_discovery(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.core.spm_controller_discovery'


class RootCoreI2c(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.core.i2c'


class RootCoreCamera(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.core.camera'


class RootCore(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.core'
        self.camera = RootCoreCamera(self._context)
        self.i2c = RootCoreI2c(self._context)
        self.spm_controller_discovery = RootCoreSpm_controller_discovery(self._context)
        self.storage = RootCoreStorage(self._context)
        self.hw_modules = RootCoreHw_modules(self._context)
        self.script_server = RootCoreScript_server(self._context)
        self.vshi = RootCoreVshi(self._context)
        self.core_options_store = RootCoreCore_options_store(self._context)


class RootTestTabel(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.test.tabel'


class RootTest(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.test'
        self.tabel = RootTestTabel(self._context)

    def func(self, *args) -> Any:
        return self._context.call('root.test.func', *args)

    @property
    def bool(self) -> bool:
        return bool(self._context.get('root.test.bool'))

    @bool.setter
    def bool(self, new_val:bool):
        self._context.set('root.test.bool', bool(new_val))

    @property
    def str(self) -> str:
        return str(self._context.get('root.test.str'))

    @str.setter
    def str(self, new_val:str):
        self._context.set('root.test.str', str(new_val))

    @property
    def num_i(self) -> int:
        return int(self._context.get('root.test.num_i'))

    @num_i.setter
    def num_i(self, new_val:int):
        self._context.set('root.test.num_i', int(new_val))

    @property
    def num_f(self) -> float:
        return float(self._context.get('root.test.num_f'))

    @num_f.setter
    def num_f(self, new_val:float):
        self._context.set('root.test.num_f', float(new_val))


class RootWorkflowManager(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.workflow.manager'

    @property
    def session_name(self) -> str:
        return str(self._context.get('root.workflow.manager.session_name'))

    @session_name.setter
    def session_name(self, new_val:str):
        self._context.set('root.workflow.manager.session_name', str(new_val))


class RootWorkflowWorkflow_log_buffer(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.workflow.workflow_log_buffer'


class RootWorkflowCantilever_browser(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.workflow.cantilever_browser'


class RootWorkflowHardware_detection(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.workflow.hardware_detection'


class RootWorkflowCamera(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.workflow.camera'


class RootWorkflow(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.workflow'
        self.camera = RootWorkflowCamera(self._context)
        self.hardware_detection = RootWorkflowHardware_detection(self._context)
        self.cantilever_browser = RootWorkflowCantilever_browser(self._context)
        self.workflow_log_buffer = RootWorkflowWorkflow_log_buffer(self._context)
        self.manager = RootWorkflowManager(self._context)


class RootSession(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root.session'

    @property
    def current_connection(self) -> str:
        return str(self._context.get('root.session.current_connection'))

    @current_connection.setter
    def current_connection(self, new_val:str):
        self._context.set('root.session.current_connection', str(new_val))

    def select_main(self, *args) -> Any:
        return self._context.call('root.session.select_main', *args)

    @property
    def name(self) -> str:
        return str(self._context.get('root.session.name'))

    @name.setter
    def name(self, new_val:str):
        self._context.set('root.session.name', str(new_val))

    def list(self, *args) -> Any:
        return self._context.call('root.session.list', *args)

    def select(self, *args) -> Any:
        return self._context.call('root.session.select', *args)


class Root(wrap.CmdTreeNode):
    def __init__(self, context: 'StudioScriptContext'):
        super().__init__()
        self._context = context
        self._lua_tree_name = 'root'
        self.session = RootSession(self._context)
        self.workflow = RootWorkflow(self._context)
        self.test = RootTest(self._context)
        self.core = RootCore(self._context)
        self.util = RootUtil(self._context)

    def log_info(self, *args) -> Any:
        return self._context.call('root.log_info', *args)

    def log_debug(self, *args) -> Any:
        return self._context.call('root.log_debug', *args)

    @property
    def init_complete(self) -> bool:
        return bool(self._context.get('root.init_complete'))

    @init_complete.setter
    def init_complete(self, new_val:bool):
        self._context.set('root.init_complete', bool(new_val))

    def log_warn(self, *args) -> Any:
        return self._context.call('root.log_warn', *args)

    def log_error(self, *args) -> Any:
        return self._context.call('root.log_error', *args)

    def log_fatal(self, *args) -> Any:
        return self._context.call('root.log_fatal', *args)






import pathlib
import enum
import xml.etree.ElementTree as xmldoc
import nanosurf.lib.devices.i2c.chip_TMC5031 as tmc

class MoveStatusFlags(enum.IntEnum): 
    Left_Switch_Active         = tmc.RampStatFlags.Stop_Left
    Right_Switch_Active        = tmc.RampStatFlags.Stop_Right
    IsNotMoving                = tmc.RampStatFlags.ZeroVelocity_Reached
    Left_Latched_Switch_Active = tmc.RampStatFlags.Stop_Left_Latched
    Right_Latched_Switch_Active = tmc.RampStatFlags.Stop_Right_Latched
    NONE = 0x00000000
    All  = (
        tmc.RampStatFlags.Stop_Left         | tmc.RampStatFlags.Stop_Right         |
        tmc.RampStatFlags.Stop_Left_Latched | tmc.RampStatFlags.Stop_Right_Latched |
        tmc.RampStatFlags.ZeroVelocity_Reached
    )
    OTHERS = 0xffffffff & (~All)

class LimitSwitchConfig(enum.IntEnum): # part of SWModeFlags
    Switch_Left_Enabled     = tmc.SWModeFlags.Switch_Left_Enabled
    Switch_Right_Enabled    = tmc.SWModeFlags.Switch_Right_Enabled
    Switch_Left_HighActive  = tmc.SWModeFlags.Switch_Left_HighActive
    Switch_Right_HighActive = tmc.SWModeFlags.Switch_Right_HighActive
    Swap_Left_Right_Switch  = tmc.SWModeFlags.Swap_Left_Right_Switch
    LatchPos_Left_Active    = tmc.SWModeFlags.LatchPos_Left_Active
    LatchPos_Left_Inactive  = tmc.SWModeFlags.LatchPos_Left_Inactive
    LatchPos_Right_Active   = tmc.SWModeFlags.LatchPos_Right_Active
    LatchPos_Right_Inactive = tmc.SWModeFlags.LatchPos_Right_Inactive
    NONE = 0x00000000
    All  = (
        tmc.SWModeFlags.Switch_Left_Enabled    | tmc.SWModeFlags.Switch_Right_Enabled    |
        tmc.SWModeFlags.Switch_Left_HighActive | tmc.SWModeFlags.Switch_Right_HighActive |
        tmc.SWModeFlags.Swap_Left_Right_Switch | 
        tmc.SWModeFlags.LatchPos_Left_Active   | tmc.SWModeFlags.LatchPos_Left_Inactive  |
        tmc.SWModeFlags.LatchPos_Right_Active  | tmc.SWModeFlags.LatchPos_Right_Inactive 
    )
    OTHERS = 0xffffffff & (~All)

class TrinamicMotorController(tmc.Chip_TMC5031):
    """ A stepper motor controller board based around the TMC3051 chip
        Multiple designs are build with different amount of motor drivers
    """
    
    #  usable for DriveAFM Approach Foot / DIMO Focus
    default_config:dict[tmc.Reg_Channel, int] = { 
        tmc.Reg_Channel.GCONF:0x08,
        tmc.Reg_Channel.X_COMPARE: 0,
        tmc.Reg_Channel.RAMPMODE: 0,
        tmc.Reg_Channel.XACTUAL: 0,
        tmc.Reg_Channel.VSTART: 0,
        tmc.Reg_Channel.A1: 32767,
        tmc.Reg_Channel.V1: 0,
        tmc.Reg_Channel.AMAX: 65535,
        tmc.Reg_Channel.VMAX: 0,
        tmc.Reg_Channel.DMAX: 65535,
        tmc.Reg_Channel.D1: 32767,
        tmc.Reg_Channel.VSTOP: 1,
        tmc.Reg_Channel.TZEROWAIT: 0,
        tmc.Reg_Channel.XTARGET: 0,
        tmc.Reg_Channel.IHOLD_IRUN: 0x00071f00,
        tmc.Reg_Channel.VCOOLTHRS: 0,
        tmc.Reg_Channel.VHIGH: 0,
        tmc.Reg_Channel.SW_MODE: 0x00000863,
        tmc.Reg_Channel.MSLUT_0: 0xaaaab554,
        tmc.Reg_Channel.MSLUT_1: 0x4a9554aa,
        tmc.Reg_Channel.MSLUT_2: 0x24492929,
        tmc.Reg_Channel.MSLUT_3: 0x10104222,
        tmc.Reg_Channel.MSLUT_4: 0xf8000000,
        tmc.Reg_Channel.MSLUT_5: 0xb5bb777d,
        tmc.Reg_Channel.MSLUT_6: 0x49295556,
        tmc.Reg_Channel.MSLUT_7: 0x80404222,
        tmc.Reg_Channel.MSLUTSEL: 0xffff9a06,
        tmc.Reg_Channel.MSLUTSTART: 0x00f80001,
        tmc.Reg_Channel.CHOPCONF: 0x101d5,
        tmc.Reg_Channel.COOLCONF: 0x0
    }

    def __init__(self, bus_addr: int, motors: int = 1, **kwargs):
        super().__init__(bus_addr, **kwargs)
        self._motors = motors

    def init_board(self, motors: int = 1, auto_config: bool = True, motor_config: dict[tmc.Reg_Channel, int] = default_config):
        self.init_spi_bridge()
        self._motors = motors
        for motor in range(self._motors):
            if auto_config:
                self.write_config(motor, motor_config) 
            self.set_current_position(0, motor)
            self.motor_driver_enable(motor, True)

    def write_config(self,  motor: int, motor_config: dict[tmc.Reg_Channel, int]):
        old_driver_state = self.is_motor_driver_enabled(motor)
        self.motor_driver_enable(motor, False)
        
        for reg, reg_val in motor_config.items():
            self.write_register(motor, reg, reg_val, signed=False)

        self.motor_driver_enable(motor, old_driver_state)

    def is_motor_moving(self, motor: int = 0) -> bool:
        reg_val = self.read_register(motor, tmc.Reg_Channel.RAMP_STAT) 
        return (reg_val & tmc.RampStatFlags.ZeroVelocity_Reached) == 0

    def is_limit_switch_active(self, motor: int = 0) -> bool:
        reg_val = self.read_register(motor, tmc.Reg_Channel.RAMP_STAT) 
        return (reg_val & (tmc.RampStatFlags.Stop_Left | tmc.RampStatFlags.Stop_Right)) > 0

    def motor_stop(self, motor: int = 0):
        self.write_register(motor, tmc.Reg_Channel.VMAX, 0)
        self.write_register(motor, tmc.Reg_Channel.XTARGET, 0)

    def get_current_position(self, motor: int = 0) -> int:
        return self.read_register(motor,tmc.Reg_Channel.XACTUAL, signed=True)

    def set_current_position(self, pos: int, motor: int = 0) -> int:
        return self.write_register(motor,tmc.Reg_Channel.XACTUAL, pos, signed=True)

    def start_move_to_absolute(self, pos: int, speed: int, motor: int = 0):
        self.write_register(motor, tmc.Reg_Channel.XTARGET, pos)
        self.write_register(motor, tmc.Reg_Channel.VMAX, speed)

    def start_move_relative(self, pos: int, speed: int, motor: int = 0):
        cur_pos = self.get_current_position(motor)
        self.write_register(motor, tmc.Reg_Channel.XTARGET, cur_pos + pos, signed=True)
        self.write_register(motor, tmc.Reg_Channel.VMAX, speed, signed=False)

    def get_move_status(self, motor: int = 0) -> MoveStatusFlags:
        stat = self.read_register(motor, tmc.Reg_Channel.RAMP_STAT)
        return stat & MoveStatusFlags.All

    def get_limit_switch_configuration(self, motor: int = 0) -> LimitSwitchConfig:
        return self.read_register(motor, tmc.Reg_Channel.SW_MODE) & LimitSwitchConfig.All

    def set_limit_switch_configuration(self, config:  LimitSwitchConfig,motor: int = 0):
        cur_val = self.read_register(motor, tmc.Reg_Channel.SW_MODE)
        self.write_register(motor, tmc.Reg_Channel.SW_MODE, (cur_val & LimitSwitchConfig.OTHERS) | (config & LimitSwitchConfig.All))
 
def from_normalized_speed(speed: float) -> int:
    """ convert normalized speed between 0 (stop) and 1.0 (max speed) to value of tmc register"""
    return int(min(max(0.0, speed), 1.0) * float((2**23)-512))

def to_normalize_speed(speed: int) -> float:
    """ convert tmc speed register value to normalized speed between 0 (stop) and 1.0 (max speed)"""
    return float(speed) / float((2**23)-512)

def from_normalized_signed_position(pos: float) -> int:
    """ convert normalized positions between -1.0 and +1.0 to value of tmc register"""
    return int(min(max(-1.0, pos), 1.0) * float((2**31)-1))

def to_normalized_signed_position(pos: int) -> float:
    """ convert tmc position regster value to normalized position between -1.0 and +1.0"""
    return float(pos) / float((2**31)-1)

def read_motor_config_from_xml(file: pathlib.Path, set_id: int) -> dict[tmc.Reg_Channel, int]:
    
    def to_value(val_str: str) -> int:
        if "0x" in val_str:
            return int(val_str[2:], base=16)
        else:
            return int(val_str)

    config_dict :dict[tmc.Reg_Channel, int]= {}
    try:
        doc = xmldoc.parse(file)
        
        # search for set to load
        found_set = None
        sets = doc.findall("set")
        for cur_set in sets:
            id = int(cur_set.attrib["id"])
            if id == set_id:
                found_set = cur_set
                break

        # if desired set was found, load all register values into a dictionary
        if found_set is not None:
            regs = found_set.findall("register")
            for reg in regs:
                config_dict[int(reg.attrib['id'])] = to_value(reg.text)
    except:
        pass
    return config_dict



""" The long lasting worker thread to perform the frequency sweep
Copyright Nanosurf AG 2021
License - MIT
"""
import time
import numpy as np
import pythoncom
from PySide2.QtCore import Signal
import nanosurf
import nanosurf.lib.spm.workflow.frequency_sweep as freq_sweep
import nanosurf.lib.datatypes.nsf_thread as nsf_thread
from app import app_common, module_base
from modules.frequency_sweep_module import sweep_settings

class FrequencySweepData():
    def __init__(self):
        self.result_ok = False
        self.result_freq:list[float] = []
        self.result_amplitude:list[float] = []
        self.result_phase:list[float] = []

class FrequencySweepWorker(nsf_thread.NSFBackgroundWorker):
    
    sig_message = Signal(str, int)
    sig_sweep_tick = Signal(float, float) # send out ticker during sweeping with remaining time

    """ parameter for the background work"""
    par_cantilever = None
    par_excitation_method : sweep_settings.ExcitationMethodID = sweep_settings.ExcitationMethodID.PiezoDrive 
    par_input_source = freq_sweep.InputSource.Deflection
    par_output_source = freq_sweep.FrequencySweepOutput.Normal_Excitation
    par_bandwidth = freq_sweep.Bandwidths.Hz_360

    par_center_frequency = 150000
    par_frequency_range = 100000
    par_frequency_step = 100
    par_excitation_amplitude = 0.2
    par_deflection_setpoint = 0

    par_plot_style_id=sweep_settings.PlotStyleID.Linear

    def __init__(self, my_module: module_base.ModuleBase):
        self.module = my_module
        self.resulting_data = FrequencySweepData()
        self.spm:nanosurf.spm.Spm = None
        super().__init__()

    def _init_in_new_thread(self):
        super()._init_in_new_thread()
        self.sig_message.connect(self.module.app.show_message)

    def send_message(self, msg:str, msg_type : app_common.MsgType = app_common.MsgType.Info):
        self.sig_message.emit(msg, msg_type)
        self.logger.info(msg)       

    def get_result(self) -> FrequencySweepData:
        return self.resulting_data

    def do_work(self):
        """ This is the working function for the long task"""
        self.resulting_data = FrequencySweepData()
        if self._connect_to_controller():
            self.application = self.spm.application
            self.freq_sweeper = freq_sweep.FrequencySweep(self.spm)
            if self.par_excitation_method == sweep_settings.ExcitationMethodID.PiezoDrive:
                self.application.OperatingMode.ExcitationMode = nanosurf.Spm.ExcitationMode.PhotoThermal 
            else:
                self.application.OperatingMode.ExcitationMode = nanosurf.Spm.ExcitationMode.PiezoElectric

            sweep_time = self.freq_sweeper.start_execute(
                start_frequency=self.par_center_frequency-(self.par_frequency_range/2),
                end_frequency=self.par_center_frequency+(self.par_frequency_range/2),
                frequency_step=self.par_frequency_step,
                sweep_amplitude=self.par_excitation_amplitude,
                input_source=self.par_input_source,
                input_range=freq_sweep.InputRanges.Full,
                mixer_bw_select=self.par_bandwidth,
                reference_phase=0.0, 
                output=self.par_output_source
                )

            print(f"Wait for {sweep_time:.1f}s.")
            start_time = time.time()
            while self.freq_sweeper.is_executing() and not self.is_stop_request_pending():
                time.sleep(0.1)
                cur_freq = self.freq_sweeper.get_current_sweep_frequency()
                remaining_time = sweep_time - (time.time() - start_time)
                self.sig_sweep_tick.emit(cur_freq, remaining_time)
            data = self.freq_sweeper.finish_execution()

            if not self.is_stop_request_pending():
                self.resulting_data.result_amplitude = np.abs(data[0])
                self.resulting_data.result_phase = np.unwrap(np.angle(data[0], deg=True), discont=180)
                self.resulting_data.result_freq = data[1]
                self.resulting_data.result_ok = True
            else:
                self.resulting_data.result_ok = False

        self._disconnect_from_controller()

    def get_result(self) -> FrequencySweepData:
        return self.resulting_data

    def _connect_to_controller(self) -> bool:
        self.send_message("Connecting to Nanosurf controller")
        ok = False
        if self.spm is None:
            pythoncom.CoInitialize()
            self.spm = nanosurf.SPM()
            if self.spm.is_connected():
                if self.spm.is_scripting_enabled():
                    ok = True
                else:
                    self.send_message("Error: Scripting interface is not enabled", app_common.MsgType.Error)
            else:
                self.send_message("Error: Could not connect to controller. Check if software is started", app_common.MsgType.Error)
        else:
            ok = True
        return ok

    def _disconnect_from_controller(self):
        if self.spm is not None:
            if self.spm.application is not None:
                del self.spm
            self.spm = None
 


"""Module for scripting the logical units of the low level scripting interface.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import enum
import nanosurf.lib.spm.lowlevel.ctrlunits.factory as ctrl_factory

class Lowlevel():
    ctrlunits = ctrl_factory._CtrlUnitFactory()

    """Contains the low-level interface classes."""
    def __init__(self, spm_ctrl_manager = None, lu_shared_file_path: str = ""):
        """Creates the objects of the low-level interface.

        Parameters
        ----------
        spm_ctrl_manager:
            COM object `application.SpmCtrlManager`

        lu_shared_file_path: str
            The file path of the file that describes the logical unit
            interface. (Usually `LogicalUnit_InterfaceShared.h`)
        """
        if spm_ctrl_manager is not None:
            self._logical_unit_com = spm_ctrl_manager.LogicalUnit
            self._create_data_buffer_interface(spm_ctrl_manager.DataBuffer)

            # Try to auto detect the LU Interface declaration
            if lu_shared_file_path == "":
                try:
                    # only v3.10.1 or newer has this property
                    lu_shared_file_path = self._logical_unit_com.GetInterfaceDescriptionFile
                except:
                    pass
            if lu_shared_file_path != "":
                self.create_logical_unit_interface(lu_shared_file_path)

    def _create_data_buffer_interface(self, data_buffer_com):
        import nanosurf.lib.spm.lowlevel.data_buffer_interface as nsf_data
        type_dict = {'_data_buffer_com': data_buffer_com}
        Interface = type(
            'DataBuffer', (nsf_data.DataBufferInterface,), type_dict)
        setattr(self, 'DataBuffer', Interface)

    def create_logical_unit_interface(self, lu_shared_file_path):
        import nanosurf.lib.spm.lowlevel.logical_unit_interface as lu
        import nanosurf.lib.spm.lowlevel.logical_unit_interface_parser as lu_parse
        
        type_items = lu_parse.ParseShared(
            lu_shared_file_path).lu_type_definitions.items()
        for type_name, type_definition in type_items:
            Interface = lu.logical_unit_type(
                    type_name, self._logical_unit_com, type_definition)
            setattr(self, type_name, Interface)

class StartStopMaskBit(enum.IntEnum): 
  DataCapture                 = 0x00000001
  DataSampling                = 0x00000002
  Unused                      = 0x00000004
  RampGenPOSITIONX            = 0x00000008
  RampGenPOSITIONY            = 0x00000010
  RampGenPLANEZ               = 0x00000020
  RampGenCTRLZ                = 0x00000040
  RampGenMAXZ                 = 0x00000080
  RampGenALTERNATEZ           = 0x00000100
  RampGenTIPVOLTAGE           = 0x00000200
  RampGenAPPROACH             = 0x00000400
  TimerPROC0                  = 0x00000800
  MemSigGen                   = 0x00001000
  RampGenTEST                 = 0x00002000
  RampGenUSER4                = 0x00004000
  TimerDBG1                   = 0x00008000

class EventMaskBits(enum.IntEnum):
    DataCapture_Done            = 0x00000001
    DataSampling_Done           = 0x00000002
    Unused_Done                 = 0x00000004
    RampGenPOSITIONX_Done       = 0x00000008
    RampGenPOSITIONY_Done       = 0x00000010
    RampGenPLANEZ_Done          = 0x00000020
    RampGenCTRLZ_Done           = 0x00000040
    RampGenMAXZ_Done            = 0x00000080
    RampGenALTERNATEZ_Done      = 0x00000100
    RampGenTIPVOLTAGE_Done      = 0x00000200
    RampGenAPPROACH_Done        = 0x00000400
    TimerPROC0_Done             = 0x00000800
    MemSigGen_Done              = 0x00001000
    RampGenTEST_Done            = 0x00002000
    RampGenUSER4_Done           = 0x00004000
    TimerDBG1_Done              = 0x00008000
    ZCtrlReachedErrorLimit_True = 0x00040000
    ZCtrlReachedMinZ_True       = 0x00080000
    ZCtrlReachedMaxZ_True       = 0x00100000
    ZCtrlReachedOutLimit_True   = 0x00200000
    ExtEvent0_True              = 0x00400000
    ExtEvent1_True              = 0x00800000
    SoftEvent0_True             = 0x01000000
    SoftEvent1_True             = 0x02000000
    SoftEvent2_True             = 0x04000000
    UserAbortEvent_True         = 0x08000000
    ANDMask0_True               = 0x10000000
    ANDMask1_True               = 0x20000000
    ANDMask2_True               = 0x40000000
    ANDMask3_True               = 0x80000000

if __name__ == '__main__':
    import nanosurf.lib.spm.lowlevel.manager_mock

    lowlevel = Lowlevel(
        nanosurf.lib.spm.lowlevel.manager_mock.SPMCtrlManager(),
        r"..\..\test\LogicalUnit_InterfaceShared.h")




"""Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""


from nanosurf.lib.devices.i2c.bus_master import I2CBusMaster, I2CBusID
from nanosurf.lib.devices.i2c.chip_24LC32A import Chip_24LC32A
from nanosurf.lib.devices.i2c.chip_PCA9534 import Chip_PCA9534
from nanosurf.lib.devices.i2c.chip_TCN75 import Chip_TCN75
from nanosurf.lib.devices.i2c.chip_TMC5031 import Chip_TMC5031
from nanosurf.lib.devices.i2c.chip_TMP42x import Chip_TMP42X




""" Property Library.  Usefull to organize, manipulate, store and change values 
Copyright Nanosurf AG 2021
License - MIT
"""
from os import read
import pathlib
import configparser
from typing import Any
from PySide2 import QtCore
from nanosurf.lib.datatypes import sci_val 

convert_str_to_type_map = {
   "str" : lambda x : repr_str_to_str(str(x)),
   "float" : lambda x : float(x),
   "int" : lambda x : int(x),
   "bool" : lambda x : bool((x == "True") or (x == "true")),
   "SciVal" : lambda x : sci_val.from_str(x),
   "WindowsPath" : lambda x : repr_str_to_pathlib(str(x)),
}

def repr_str_to_str(repstr:str) -> str:
    return repstr.removeprefix("'").removesuffix("'")

def repr_str_to_pathlib(path_str: str) -> pathlib.Path:
    return pathlib.Path(path_str.removeprefix("WindowsPath('").removesuffix("')"))

class PropVal(QtCore.QObject):
    sig_value_changed = QtCore.Signal()
    
    def __init__(self, var = None, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.__var = var

    # property based interface

    @property
    def var(self) -> Any:
        return self.__var

    @property
    def value(self) -> Any:
        if isinstance(self.__var, sci_val.SciVal):
            return self.__var.value()
        else:
            return self.__var

    @value.setter
    def value(self, val):
        if isinstance(self.__var, sci_val.SciVal) :
            if isinstance(val, sci_val.SciVal): 
                self.__var = val
                self.sig_value_changed.emit()        
            else:
                if  self.__var.value() != val:
                    self.__var.set_value(val)
                    self.sig_value_changed.emit()
        else:
            if isinstance(val, sci_val.SciVal): 
                if self.__var != val.value():
                    self.__var = val.value()
                    self.sig_value_changed.emit()
            else:
                assert type(val) == type(self.__var), "Error: Type of argument has to be equal to initial value type. Use change_type() instead"
                if self.__var != val:
                    self.__var = val
                    self.sig_value_changed.emit()

    # classic style interface. 
    # For calls where the property based interface cannot be used (e.g as lambda function)
    def set_value(self, val):
        self.value = val

    def get_value(self):
        return self.value

    def change_type(self, new_value:Any):
        self.__var = new_value
        self.sig_value_changed.emit()

    def serialize(self) -> str:
        return type(self.__var).__qualname__ + "@" + repr(self.__var)

    def deserialize(self, str) -> bool:
        type_str, ser_val = str.split("@")
        done = False
        try:
            self.value = convert_str_to_type_map[type_str](ser_val)
            done = True
        except:
            self.__var = None
        return done

class PropStore():
    def has_property(self, prop_name: str) ->bool:
        prop = getattr(self, prop_name, None)
        return isinstance(prop, PropVal)

    def get_property(self, prop_name: str) -> PropVal:
        prop = getattr(self, prop_name, None)
        if isinstance(prop, PropVal):
            return prop
        else:
            return PropVal()

    def set_property(self, prop_name: str, prop_value: PropVal):
        setattr(self, prop_name, prop_value)

    def get_property_dict(self, hide_locals: bool = True) -> dict:
        prop_dict = {}
        for v in dir(self):
            if isinstance(getattr(self, v), PropVal):
                if not(hide_locals and v[0] == "_"): 
                    prop_dict[v] = getattr(self, v)
        return prop_dict

    def serialize(self, ignore_locals: bool = True) -> dict:
        prop_dict = self.get_property_dict(hide_locals=ignore_locals)
        
        name_serial_str_dict = {}
        try:
            for name, prop in prop_dict.items():
                name_serial_str_dict[name] = prop.serialize()
        except:
            pass
        return name_serial_str_dict

    def deserialize(self, name_serial_str_dict : dict, ignore_locals: bool = True) -> bool:
        # convert each key into a property
        done = True
        for name, ser_str in name_serial_str_dict.items():
            if not (name[0]=='_' and ignore_locals):
                # create new or update property
                prop = getattr(self, name, PropVal())
                if prop.deserialize(ser_str):
                    self.set_property(name, prop)
                else:
                    done = False
        return done

def save_to_ini_file(store: PropStore, file: pathlib.Path, section: str, 
    do_not_save_locals: bool = True, update_content: bool = True) -> bool:
    """ Save all properties and their values into a text file in a human readable form
        
    Parameters
    ----------
    file
        The file to store into. Including path 
    do_not_save_locals
        if this option is Tue, then do not save properties with a "_" prefix in their name

    Returns
    -------
    bool 
        True in case the write to file was successful, otherwise False
    """

    try:
        config = configparser.ConfigParser()
        config.optionxform = lambda option: option # preserve upper/lower case

        if update_content:
            config.read(file)

        key_value_dict = store.serialize(ignore_locals=do_not_save_locals)
        config[section] = key_value_dict

        with open(file, 'w') as f:
            config.write(f)
            f.close()
        done = True
    except:
        done = False        
    return done

def load_from_ini_file(store:PropStore, file: pathlib.Path, section: str, 
    do_not_load_locals: bool = True) -> bool:
    """ Load all properties and their values from a text file
    
    Parameters
    ----------
    file
        The file to store into. Including path 
    do_not_load_locals
        if this option is Tue, then do not load properties with a "_" prefix in their name

    Returns
    -------
    bool 
        True in case the write to file was successful, otherwise False
    """
    done = False
    try:
        # read in section
        config = configparser.ConfigParser(interpolation=None)
        config.optionxform = lambda option: option # preserve upper/lower case
        config.read(file)

        sec = config[section]

        store.deserialize(sec, ignore_locals=do_not_load_locals)
        done = True
    except:
        done = False
    return done



""" A base class for background tasks
Copyright Nanosurf AG 2021
License - MIT
"""
import sys
import time
import logging
from PySide2.QtCore import QThread, QObject, Signal
from app import module_base
from app import app_common

class WorkerData():
    """ Use this class as base class for own worker thread results"""
    pass

class SingleRunWorker(QObject):
    """ This class implements a long lasting activity in the background to not freeze the gui 
        (e.g. measuring some values over time) background 
        Subclass from this and implement do_work(). 
        
        The main thread then calls thread.start(). That's it.
        Signals emitted from worker thread to inform about its state
        Signal 'sig_finish' tells you when the work is done.

        The sub class can fire 'sig_new_data' to inform about the availability of new data.
        client can then call get_data() to read this.

        Text messages can be sent to the application by 'sig_message' any time.

        A running task can be asked if it's still running by task.is_running()
        A request to stop is done by: task.stop(wait=True|False)
        
        The user function in task.do_work() should implement the possibility to abort the work by checking regularly self.stop_request_flag 
        To check wether a task was aborted call: task.is_aborted()
    """
    sig_started = Signal()
    sig_new_data = Signal()
    sig_finished = Signal()

    def start(self):
        """ Start the worker activity. 
        'sig_started' is emitted to inform the real start of the worker activity"""
        if not self.is_running():
            self.stop_request_flag = False
            self.thread.start()
        else:
            self.logger.warning("Thread allready started")

    def stop(self, wait: bool = False, timeout : float = 10.0) -> bool:
        """ Send stop request to worker thread.  
            'sig_finished' is emitted when the worker thread has finished its job.

            Parameters
            ----------
            wait: bool
                If True the function waits until the worker thread has really finished
            timeout: float
                maximal wait time 

            Returns
            -------
            Bool:
                True if task could be stopped.
        """
        if self.is_running():
            self.stop_request_flag = True

            if wait:
                abort_waiting = False
                ticks = 0
                waiting_time = 0.0
                while self.is_running() and not abort_waiting:
                    ticks += 1
                    if ticks > 10:
                        waiting_time += 1.0
                        if waiting_time > timeout:
                            abort_waiting = True
                            self.logger.warning(f"Stop request was not served in {timeout}s")
                            self.thread.quit()
                    time.sleep(0.1)
        else:
            self.logger.warning("Thread allready stopped")
        return not self.is_running()

    def is_running(self) -> bool:
        return self.is_running_flag 

    def is_aborted(self) -> bool:
        return self.stop_request_flag and (not self.is_running_flag)

    def get_result(self) -> WorkerData:
        return self.resulting_data

    #--------------- for sub class only ------------------------------------------------------------

    def send_message(self, msg:str, msg_type : app_common.MsgType = app_common.MsgType.Info):
        self._sig_message.emit(msg, msg_type)
        self.logger.info(msg)       

    def do_work(self):
        """ This function has to be overwritten by sub class and do the long lasting background work"""
        raise NotImplementedError(f"Subclass of '{self.__class__.__name__}' has to implement '{sys._getframe().f_code.co_name}()'")

    #--------------- internal functions ------------------------------------------------------------

    _sig_ready_to_quit = Signal() # internal use
    _sig_message = Signal(str, int)
    
    def __init__(self, my_module: module_base.ModuleBase):
        """ setup the thread and wait until the task is started by thread.start()"""
        super().__init__()
        self.module = my_module
        self.logger = logging.getLogger(self.module.name)
        self.stop_request_flag = False
        self.is_running_flag = False
        self.resulting_data = WorkerData()
        self.thread = QThread()
        self._sig_ready_to_quit.connect(self.thread.quit)
        self._sig_message.connect(self.module.app.show_message)
        
        self.moveToThread(self.thread)
        self.thread.started.connect(self._background_worker)

    def _background_worker(self):
        """ this function is called at task start""" 

        # prepare 
        self.is_running_flag = True
        self.stop_request_flag = False
        self.module.app.activate_debugger_support_for_this_thread()
        self.sig_started.emit()

        # do the real stuff now
        self.do_work()
        
        # clean up
        self.is_running_flag = False
        self.sig_finished.emit()
        self._sig_ready_to_quit.emit()

